#!/usr/bin/perl

# --------------------------------------------------------------
# Prontus CMS
# http://www.prontus.cl
# by Altavoz.net
#
# licensed under LGPL license.
# http://www.prontus.cl/license.html
# --------------------------------------------------------------

# El Formulario Prontus (prontus_form.cgi) es una aplicacion cgi que permite procesar el contenido
# de un formulario publicado a traves de un articulo Prontus.
# De esta manera, el operador Prontus crear formularios de contacto, encuestas y otros a voluntad,
# tal como lo hace con cualquier otro articulo Prontus.
# Al recibir como parametros ocultos las variables _ts (timestamp), _prontus_id y _file,
# prontus_form lee desde el xml del articulo todos los datos que requiere para el procesamiento
# del formulario.
#
# Los Formularios Prontus funcionan tomando datos publicados como parte del articulo y en base
# a ellos envia y opcionalmente almacena la informacion (respaldo de datos).
#
# El proceso de un Formulario Prontus se realiza mediante los siguientes pasos:
#
# 1.Se leen los datos submitidos y se ubica el articulo Prontus correspondiente.
# 2.Se leen los datos del articulo Prontus.
# 3.Se validan los datos y se leen las plantillas de respuesta.
# 4.Se construye la pagina de respuesta unica para el visitante y se envian los correos electronicos al admin.
# Nota: la autorrespuesta está deshabilitada. 2020-11-05 avaroli@altavoz.net
# 5.Si es pertinente, se respaldan los datos.
# 6.Se redirige el browser hacia la pagina de respuesta a traves de un header Location.
# 7.Se limpia el directorio de paginas de respuesta eliminando las que tienen mas de 10 minutos de antigüedad.
#
# Configuracion
# -------------
# El Formulario Prontus se configura mediante variables Prontus estandar, a las que se agregan
# otras de uso reservado.
#
# * Variables Prontus Estandar
#   _ts          Timestamp de Prontus.
#   _file        Path completo al articulo Prontus.
#   _prontus_id  Identificador de la instancia de Prontus.
#   _form_vista  Identifica la vista que debe usarse para el proceso.
#                Si no existe esta variable, se asume la vista por defecto.
#
# * Variables Prontus Reservadas
# Para que el Formulario Prontus funcione correctamente, existe una serie de variables
# Prontus que deben definirse. Estas variables seran almacenadas en el archivo XML
# correspondiente, siendo leidas desde el por la cgi prontus_form.
#
# form_admin           E-mails de destino del formulario, separadas por comas.
#                      Si como parte del formulario se envia un dato con el nombre _admin,
#                      y es un numero, entonces el formulario se enviara al mail
#                      que esta en ese lugar de la lista separada por comas, partiendo de 1.
#                      Si el número no se encuentra en la lista, se envia a todos los emails en form_admin.
# form_from            E-mail del remitente. Esta variable es sustituida por el dato email,
#                      si es que existe en el formulario.
#                      Si el dato email no existe en el formulario, entonces no se enviara un
#                      mail de autorrespuesta.
#                      Nota: la autorrespuesta está deshabilitada. 2020-11-05 avaroli@altavoz.net
#
# form_subject<_vista> Subject (asunto) del mensaje que llegara al administrador.
# form_subject_auto<_vista> Subject (asunto) del mensaje de autorrespuesta que se envia al remitente.
# form_signature<_vista> Firma del mail de salida
# form_msg_auto<_vista>  Mensaje de autorrespuesta que se envia al remitente.
# form_msg_exito<_vista> Mensaje de exito del sistema, en formato html.
# form_msg_error<_vista> Mensaje de error del sistema, en formato html.
#
# CHK_form_required_<nombre>   Indica que el dato <nombre> es obligatorio en el formulario.
# CHK_form_captcha_enable      Indica si se validará el captcha o no
# CHK_form_backup_datos        Si existe, realiza una copia de los datos recibidos en
#                                el directorio de respaldo, incluyendo los archivos adjuntos.
# Validacion de datos
# -------------------
# El Formulario Prontus validara automaticamente los datos que comiencen con:
#   - email o e-mail (se asume que es una direccion de correo electronico)
#   - rut            (se asume que es un RUT)
#   - telefono, phone, fax, celular o cellphone (se asume que es un numero telefonico)
#
# Los datos que comiencen con un underscore (_) no seran enviados ni almacenados.
#
# Plantillas
# ----------
# Las plantillas a utilizar se especifican en el formulario mediante los datos:
# _pag_exito = Plantilla de exito del sistema (solo el nombre del archivo, sin extension).
# _pag_error = Plantilla de error del sistema (solo el nombre del archivo, sin extension).
#
# Si estos datos no son incluidos, el Formulario Prontus empleara las plantillas:
#
# /<prontus_id>/plantillas/extra/form/pags/exito.<ext> - Plantilla de exito y autorespuesta por defecto.
# /<prontus_id>/plantillas/extra/form/pags/error.<ext> - Plantilla de error por defecto.
#
# La extension debe ser la misma usada en el formulario, ya que se deduce de este.
#
# Dentro de ellos, la marca %%MSG%% sera sustituida por el mensaje del sistema.
#
# Servidor SMTP
# -------------
# El Formulario Prontus utiliza el servidor SMTP definido en la configuración.
#
# Archivos temporales
# -------------------
# Los archivos temporales utilizados para entregar la respuesta al visitante se encuentran en:
#
# /<prontus_id>/site/cache/form/pags
#
# To-Do:
# - Revisar sensibilidad a las mayusculas.

# -------------------------------BEGIN SCRIPT--------------------

use strict;
use utf8;

BEGIN {
    use FindBin qw($Bin);
    my $pathLibs = $Bin;
    unshift(@INC, $pathLibs);

    $pathLibs =~ s/\/[^\/]+$/\/cgi\-cpn/;
    unshift(@INC,$pathLibs);
}

##libs##
use lib_stdlog;
lib_stdlog::set_stdlog($0, 51200);
use prontus_varglb;
prontus_varglb::init();

use Digest::SHA;
use Email::MIME;
use Email::Sender::Simple;
use Email::Sender::Transport::SMTP;
use Email::Simple;
use Encode;
use File::Copy;
use File::LibMagic;
use IO::All;
use JSON;
use LWP::UserAgent;
use POSIX;
use glib_cgi_05;
use glib_fildir_03;
use glib_hrfec_03;
use glib_html_03;
use glib_str_03;
use lib_captcha2;
use lib_encodes;
use lib_form;
use lib_lang;
use lib_mail;
use lib_maxrunning;
use lib_prontus;
use lib_utils;
use lib_validator;
use mime_types;
##/libs##

# Flag para debug.
our $DEBUG = 0;

# Directorios de trabajo, dentro de la instancia de Prontus.
our $DATA_DIR = 'cpan/procs/form'; # Directorio para el almacenamiento de los datos de respaldo.
our $TMP_DIR   = 'plantillas/extra/form'; # /pags-vvv Directorio de las plantillas de exito y fracaso.
our $CACHE_DIR = 'site/cache/form'; # /pags-vvv Directorio de las paginas generadas.
our $VISTADIR = ''; # Vista a utilizar en directorios.
our $VISTAVAR = ''; # Sufijo de vista para variables.
our $SEPARADOR = ';'; # Separador de campos del formato csv.
our $TS = '';         # Timestamp del articulo que contiene al formulario.
our $PRONTUS_ID = '';   # Nombre del Prontus.
our %PRONTUS_VARS = (); # Variables del xml Prontus.
our @DATOS = ();        # Nombres de los datos del formulario.
our @DATOS4IF = ();     # Nombres de los datos del formulario que vienen completos. Usado para procesar los IFs.
our @ADMIN_MAILS = ();  # E-Mails de administracion (donde hay que dirigir la data del formulario).
our $TMP_EXITO = '';    # Plantilla de exito.
our $TMP_ERROR = '';    # Plantilla de error.
our $ANSWERS_DIR = '';  # Directorio de las paginas de respuesta.
our $ANSWERID = '';     # Identificacion de la respuesta.
our $FECHA = '';        # Fecha de creacion del formulario. Usada para acceder a los datos xml.
our $EXT = '';          # Extension usada por Prontus. Se deduce del path del formulario.
our %MSGS = ();         # Mensajes de error.
our $BUFFER_ART = '';   # Buffer para cargar el contenido de articulo
our $SALIDA_ESTATICA = '';
our $NOM_PLT_EXITO = '';
our $NOM_PLT_ERROR = '';
our $RECAPTCHA_RESPONSE = ""; # Captura la respuesta por POST del formulario

# Variables globales.
our $ROOTDIR = $prontus_varglb::DIR_SERVER;
our %RESERVED_DATA;

main: {
    lib_utils::setUp();
    glib_cgi_05::new();

    # Soporta un maximo de n copias corriendo.
    if (lib_maxrunning::maxExcedido(10)) {
        lib_form::aborta(lib_lang('global_error_server_busy_retry'));
    }

    %PRONTUS_VARS = ();
    %RESERVED_DATA = ();

    # Inicializaciones varias.
    # Inicializa las variables de invocacion.
    &parse_request_data();

    if ($PRONTUS_VARS{'_alta'} ne '1') {
        lib_form::aborta(lib_lang::get_interpolated_message("form_error_init"));
    }

    # Carga variables de configuracion de prontus.
    my $msg_load = lib_utils::load_config_by_id($PRONTUS_ID);
    if ($msg_load ne '' ) {
        lib_form::aborta(lib_lang::get_interpolated_message("global_error_loading_config", $msg_load));
    }

    # Inicializacion de mensajes de error.
    $MSGS{'out_of_service'} = lib_lang::get_message('form_out_of_service');
    $MSGS{'required_data'} = lib_lang::get_message('form_required_data');
    $MSGS{'wrong_data'} = lib_lang::get_message('form_wrong_data');
    $MSGS{'wrong_captcha'} = lib_lang::get_message('form_wrong_captcha');
    $MSGS{'wrong_vista'} = lib_lang::get_message('form_wrong_vista');
    $MSGS{'wrong_google_captcha'} = lib_lang::get_message('form_wrong_google_captcha');
    $MSGS{'invalid_file'} = lib_lang::get_message('form_invalid_file');

    $MSGS{'nombre'} = lib_lang::get_message('form_nombre');
    $MSGS{'apellidos'} = lib_lang::get_message('form_apellidos');
    $MSGS{'email'} = lib_lang::get_message('form_email');
    $MSGS{'telefono'} = lib_lang::get_message('form_telefono');
    $MSGS{'celular'} = lib_lang::get_message('form_celular');
    $MSGS{'fax'} = lib_lang::get_message('form_fax');
    $MSGS{'rut'} = lib_lang::get_message('form_rut');

    if ($DEBUG) {
        print "Content-type: text/html\n";
        print "\n<html><body><pre>Debug Mode\n";
    }

    # Valida las variables Prontus y las del formulario.
    &valida_data();

    # Envia mails y guarda datos si es pertinente.
    my $result = &data_management();
    if ($DEBUG) { print $result; };

    # Finaliza mostrando la pagina de exito.
    &salida('', $PRONTUS_VARS{'form_msg_exito'.$VISTAVAR}, $TMP_EXITO, 0);
}
#----------------------------------------------------
# Subrutinas
#----------------------------------------------------
# Envia mails y guarda datos si es pertinente.
sub data_management {
    my($body,$data,$backupdir,$backupdata,$backupheaders);
    my ($to,$from,$subj,$filename,$filedata,$fecha,$hora,$ip);
    my ($result);
    my (%files);
    my $random = time . $$. int(rand(100000000)); # Usa el numero de proceso y un random
    # Detecta informacion de fecha, hora e IP.
    $fecha = glib_hrfec_03::fecha_human();
    $hora = glib_hrfec_03::hora_human();

    if (lib_prontus::cache_active() && defined($ENV{'HTTP_X_FORWARDED_FOR'})) {
        $ip = $ENV{'HTTP_X_FORWARDED_FOR'};
    } else {
        $ip = $ENV{'REMOTE_ADDR'};
    }

    my $plt_buffer = "";
    my $plt_loop;
    my $plt_loop_parsed;

    if ($PRONTUS_VARS{'cmb_form_admin_mail_plt'}) {
        my $plt_file = "$ROOTDIR/$PRONTUS_ID/$TMP_DIR/email/$PRONTUS_VARS{'_fid'}/$PRONTUS_VARS{'cmb_form_admin_mail_plt'}";
        # lib_stdlog::print_err("plt_file[$plt_file]\n");

        if (-f $plt_file) {
            $plt_buffer = glib_fildir_03::read_file($plt_file);

            if ($plt_buffer =~ /%%loop%%(.*?)%%\/loop%%/s) {
                $plt_loop = $1;
            }
        }
    }

    $backupdir = "$ROOTDIR/$PRONTUS_ID/$DATA_DIR/$TS";
    glib_fildir_03::check_dir($backupdir);
    # Se obtiene el TS del envío
    my $TSENVIO = glib_hrfec_03::get_dtime_pack4();
    while(-f $backupdir.'/'.$TSENVIO.'.json') {
        $TSENVIO = glib_hrfec_03::suma_segs($TSENVIO, 1);
    }

    my $data_json;
    $data_json->{'_fecha'} = $fecha;
    $data_json->{'_hora'} = $hora;
    $data_json->{'_ip'} = $ip;

    my $file_final_path = '';
    my $order_data;
    my $counter = 1;
    # Forma cuerpo para el administrador.
    # Lee formulario para determinar el orden de los datos y descartar datos inyectados.
    $BUFFER_ART = glib_fildir_03::read_file($ROOTDIR.$RESERVED_DATA{'_file'});

    my @exts = ();
    if (defined($PRONTUS_VARS{'form_ext'})) {
        @exts = split(/,/, $PRONTUS_VARS{'form_ext'});
        # es posible que solo sea una extension
        if (scalar @exts < 1) {
            push(@exts, $PRONTUS_VARS{'form_ext'});
        }
    }
    $body = lib_lang::get_message('global_msg_log_header')."\n\n";
    $backupheaders .= "\"".lib_lang::get_message('form_fecha')."\"$SEPARADOR\"".lib_lang::get_message('form_hora')."\"$SEPARADOR\"".lib_lang::get_message('form_ip')."\"$SEPARADOR";
    $backupdata .= "\"$fecha\"$SEPARADOR\"$hora\"$SEPARADOR\"$ip\"$SEPARADOR";
    foreach my $key (sort {sort_form_data($a,$b)} @DATOS) {
        # campos que se deben omitir de los datos recibidos ya que no son datos
        next if (substr($key, 0, 1) eq '_' || substr($key, 0, 17) eq 'chk_form_required' || substr($key, 0, 21) eq 'chk_form_backup_datos' || substr($key, 0, 11) eq 'g-recaptcha');
        $backupheaders .= '"' . $key . '"' . $SEPARADOR;
        # Determino si el dato es un archivo o no.
        # Si son varios, se adjuntara el ultimo.
        my $aux = glib_cgi_05::real_paths($key);
        if (defined($aux) && $aux ne '') { # Es un archivo.
            $filename = $aux;
            $filename =~ s/.+[\/\\]([^\/\\]+)/$1/; # Extrae path por si lo trae.

            $filedata = glib_cgi_05::param($key);

            my $ext = '';
            $filename =~ /(.*)\.(\w+)$/;
            $filename = $1;
            $ext = lc $2;

            # verifica extension valida
            if (scalar @exts) {
                my $type = `file -b --mime-type $filedata`;
                $type =~ s/\s//g;
                my $is_valid = 0;
                foreach my $valid_ext (@exts) {
                    if ($valid_ext eq $ext) {
                        $is_valid = 1;
                        last;
                    }
                }

                # archivo invalido o tipo mime no coincide con la extension del archivo
                if (!$is_valid || $type ne mime_types::ext_2_type($ext)) {
                    # lib_stdlog::print_err("[$aux][$ext] - [$type] ne [" . mime_types::ext_2_type($ext) . "]\n");
                    &salida($MSGS{'invalid_file'} . ' '.$aux,$PRONTUS_VARS{'form_msg_error'.$VISTAVAR},$TMP_ERROR,1);
                }
            }

            $ext = '.' . $ext; # ext con punto

            $filename = lib_encodes::no_accent_mark($filename);
            # Elimina entidades html.
            $filename =~ s/&[^;]+;//g;
            # convierte todos los caracteres que no son de palabras en espacios
            $filename =~ s/[^a-zA-Z0-9_\-\.]/ /g;
            $filename =~ s/ {2,}/ /g; # elimina espacios repetidos
            $filename =~ s/^\s+|\s+$//g; # elimina espacios iniciales y finales
            $filename =~ s/^-+|-+$//g;   # elimina todos los guiones al principio y al final
            $filename =~ s/ /\_/g; # convierte espacios a _

            $filename = $filename . $ext;

            $files{$key}{'_name'} = 'file_' . $random.'--'.$filename;
            $files{$key}{'_temp'} = $filedata;
            my $fileurl = "$prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'SERVER_PROTOCOLO_CPAN'}://$prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'CPAN_SERVER_NAME'}/$prontus_varglb::PRONTUS_ID/$DATA_DIR/$TS/$files{$key}{'_name'}";
            $body .= sprintf('%-15s',$key) . " = $fileurl\n";
            $file_final_path = "$prontus_varglb::DIR_SERVER/$prontus_varglb::PRONTUS_ID/$DATA_DIR/$TS/$files{$key}{'_name'}";
            $backupdata .= "\"file_$random--$filename\"$SEPARADOR"; # Pega el random para que el nombre sea unico.

            $plt_buffer =~ s/%%$key%%/$fileurl/sg if ($plt_buffer);

            if ($plt_loop) {
                my $loop_tmp = $plt_loop;

                $loop_tmp =~ s/%%_campo_nombre%%/$key/sg;
                $loop_tmp =~ s/%%_campo_valor%%/$fileurl/sg;

                $plt_loop_parsed .= $loop_tmp;
            }
        } else {
            $order_data->{$counter} = $key;
            $counter++;
            $data_json->{$key} = glib_cgi_05::param($key);
            $data = glib_str_03::trim(glib_cgi_05::param($key)); # Elimina espacios para que no molesten.
            $body .= glib_html_03::text2html(sprintf('%-15s',$key) . " = $data\n");
            $data =~ s/\"/\'\'/gs; # Convierte comillas para que no arruinen el archivo csv.
            $data =~ s/\r//gs;     # Elimina retornos de carro para que no arruinen el archivo csv.
            $backupdata .= "\"$data\"$SEPARADOR";

            $plt_buffer =~ s/%%$key%%/$data/sg if ($plt_buffer);

            if ($plt_loop) {
                my $loop_tmp = $plt_loop;

                $loop_tmp =~ s/%%_campo_nombre%%/$key/sg;
                $loop_tmp =~ s/%%_campo_valor%%/$data/sg;

                $plt_loop_parsed .= $loop_tmp;
            }
        }
    }

    if ($plt_buffer) {
        $plt_buffer =~ s/%%loop%%.*?%%\/loop%%/$plt_loop_parsed/sg if ($plt_loop);
        $plt_buffer = &procesaIFs($plt_buffer, 2);
    }

    my $json_str = JSON::encode_json($order_data);
    utf8::decode($json_str);
    glib_fildir_03::write_file("$backupdir/order.json", $json_str);

    # Sólo se incluyen los archivos en el JSON, si hay respaldo
    my $files_json;

    # Genera el backup
    if (-e "$backupdir/backup.csv") { # Si existe ya el archivo, no inserta la linea de encabezados.
        glib_fildir_03::append_file("$backupdir/backup.csv","$backupdata\r\n");
    } else {
        glib_fildir_03::append_file("$backupdir/backup.csv","$backupheaders\r\n$backupdata\r\n");
    }

    # Mueve todos los archivos adjuntos.
    foreach my $file (keys %files) {
        my $name = $files{$file}{'_name'};
        my $temp = $files{$file}{'_temp'};
        File::Copy::move($temp,"$backupdir/$name");
        my $newfile = "$backupdir/$name";
        $newfile =~ s/^$prontus_varglb::DIR_SERVER//;
        $files_json->{$file} = $newfile;
    }

    # Se escribe la respuesta json
    if ($data_json) {
        my $resp;
        foreach my $llave (keys %{$data_json}) {
            $resp->{$llave} = $data_json->{$llave};
        }
        if (keys %{$files_json}) {
            $resp->{'_files'} = $files_json;
        }

        my $filejson = "$backupdir/$TSENVIO.json";
        glib_fildir_03::check_dir($backupdir);

        my $json_str = JSON::encode_json($resp);
        utf8::decode($json_str);
        glib_fildir_03::write_file($filejson, $json_str);
    }

    my %artic_data = (
        'fid' => $PRONTUS_VARS{'_fid'},
        'url' => $PRONTUS_VARS{'_url'},
        'ts' => $TS,
        'prontus_id' => $prontus_varglb::PRONTUS_ID,
        'mv' => '',
        'ext' => 'html'
    );

    my $lnk = "$prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'SERVER_PROTOCOLO_FRONTEND'}://$prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'PUBLIC_SERVER_NAME'}" . lib_prontus::parse_filef('%%_fileurl%%', \%artic_data);

    # firma configurable
    my $msg_signature = "\r\n".lib_lang::get_interpolated_message('form_msg_received', $fecha, $hora, $ip)."\n";


    $msg_signature .= "\n". lib_lang::get_interpolated_message('global_article') .": $TS - ". lib_lang::get_interpolated_message('global_title_form').": $lnk\n";
    $msg_signature .= "\n".lib_lang::get_message('form_msg_thanks')."\r\n\r\n";
    if($PRONTUS_VARS{'form_signature'.$VISTAVAR}) {
        $msg_signature = "\r\n" . $PRONTUS_VARS{'form_signature'.$VISTAVAR};
    }

    if ($plt_buffer) {
        $plt_buffer =~ s/%%_fecha%%/$fecha/sg;
        $plt_buffer =~ s/%%_hora%%/$hora/sg;
        $plt_buffer =~ s/%%_ip%%/$ip/sg;
        $plt_buffer =~ s/%%_form_signature%%/$msg_signature/sg;
    }

    $body .= $msg_signature;

    # Envia mail a el o los administradores.
    # Remitente para el admin
    # Se agrega campo para correo remitente fijo distinto de autorespuesta
    if (defined($PRONTUS_VARS{'form_remitente'}) && $PRONTUS_VARS{'form_remitente'} ne '') {
        $from = $PRONTUS_VARS{'form_remitente'};
    } else {
        $from = $PRONTUS_VARS{'form_from'};
    }

    if (defined($PRONTUS_VARS{'form_subject'.$VISTAVAR}) && $PRONTUS_VARS{'form_subject'.$VISTAVAR} ne '') {
        $subj = $PRONTUS_VARS{'form_subject'.$VISTAVAR};
    } else {
        $subj = $PRONTUS_VARS{'form_subject'};
    }

    $subj = &procesaIFs($subj,1);
    foreach my $key (@DATOS) {
        # Elimina espacios para que no molesten.
        $data = glib_str_03::trim(glib_cgi_05::param($key));
        # Reemplaza datos en subject.
        $data =~ s/[^\w\-áéíóúüñÁÉÍÓÚÜÑ\, ]//g; # Elimina todo caracter extrano.
        $subj =~ s/\%\Q$key\E\%/$data/sg;
    }
    $subj =~ s/%\w+%//sg; # Elimina tags no parseados.
    my $replyto = glib_cgi_05::param('email');
    my $encode_html = 0;

    if ($plt_buffer) {
        $body = $plt_buffer;
        $encode_html = 1;
    }

    foreach my $email (@ADMIN_MAILS) {
        $to = $email;
        # si el adjunto no existe, no se envia por correo para evitar error
        if ($prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'FORM_INCLUIR_ADJUNTO'} eq 'NO' || ($file_final_path ne '' && !-s $file_final_path)) {
            $result .= ' 1 ' . lib_form::envia_mail($to, $from, $replyto, $subj, $body, '', '', $encode_html);
        } else {
            $result .= ' 1 ' . lib_form::envia_mail($to, $from, $replyto, $subj, $body, $filename, $file_final_path, $encode_html);
        }
    }

    return $result; # $result es solo para debug.
}

# Devuelve TRUE para aquellas variables del array DATOS que se encuentran en el BUFFER_ART.
sub campo_existe_html {
    my $campo = $_[0];
    return index($BUFFER_ART, "\"$campo\"") >= 0;
}

# ordena los campos segun los encuentra en el html
sub sort_form_data {
    my ($a, $b) = @_;
    my $index_a = index($BUFFER_ART,"\"$a\"");
    my $index_b = index($BUFFER_ART,"\"$b\"");
    if ($index_a < 0) {
        return 1;
    }
    if ($index_b < 0) {
        return -1
    }
    return $index_a <=> $index_b;
}

# Valida las variables Prontus y las del formulario.
sub valida_data {
    my ($file,$nombre,$dato,$plantilla);
    my @mails;

    @DATOS = glib_cgi_05::param();

    # Determina la vista que se usara.
    $VISTADIR = $RESERVED_DATA{'_form_vista'};
    $VISTADIR =~ s/[^a-z]//sg;
    if ($VISTADIR ne '') {
        # Variable queda lista para ser inserta en los ids de variables.
        $VISTAVAR = '_' . $VISTADIR;
        # Variable queda lista para ser inserta en los paths.
        $VISTADIR = '-' . $VISTADIR;
    }
    # Determina cuales seran los emails de destino (administrador).
    # por defecto se envía a todos
    my $form_admin = $PRONTUS_VARS{'form_admin'};
    $RESERVED_DATA{'_admin'} =~ s/[^\d]+//g;  # Elimina cualquier caracter no numerico

    # Si en el form hay un dato _admin, y es numérico, se busca ese número en la lista definida en el FID.
    # Si no existe el dato, se usa el o los emails definidos en el FID.
    if ($RESERVED_DATA{'_admin'} ne '' && $RESERVED_DATA{'_admin'} > 0) {
        my @mails = split(/,/, $PRONTUS_VARS{'form_admin'});
        # existe este correo en la lista
        if (exists($mails[$RESERVED_DATA{'_admin'} - 1])) {
            $form_admin = $mails[$RESERVED_DATA{'_admin'} - 1];
        }
    }

    if ($form_admin eq '') {
        lib_form::aborta(lib_lang::get_message('form_error_no_destination'));
    }

    # Validacion de las variables Prontus.
    @ADMIN_MAILS = split(/,/, $form_admin);
    foreach my $email (@ADMIN_MAILS) {
        if ($email !~ /[\w\.\-\,\']+\@[\w\-\.]+\.[\w]+/) {
            lib_form::aborta(lib_lang::get_message('form_error_no_destination'));
        }
    }
    # Permite que el mail del remitente sea vacio. En ese caso no se envia mail de autorrespuesta.
    # Nota: la autorrespuesta está deshabilitada. 2020-11-05 avaroli@altavoz.net
    if ($PRONTUS_VARS{'form_from'} ne '') {
        if ($PRONTUS_VARS{'form_from'} !~ /[\w\.\-\,\']+\@[\w\-\.]+\.[\w]+/) {
            lib_form::aborta(lib_lang::get_interpolated_message('form_error_wrong_from_email', $PRONTUS_VARS{'form_from'}));
        }
    }

    # Deduce la extension del formulario.
    $file = $RESERVED_DATA{'_file'};
    if ($file =~ /\.([\w]+)$/) {
        $EXT = $1;
    } else {
        lib_form::aborta(lib_lang::get_message('form_error_invalid_file'));
    }

    # Establece cuales son y verifica que existen las plantillas basicas.
    if ($RESERVED_DATA{'_pag_exito'} ne '' ) {
        $plantilla = $RESERVED_DATA{'_pag_exito'};
        $plantilla =~ s/[^\w\-]//g; # Solo caracteres alfanumericos y - y _.
        $TMP_EXITO = glib_fildir_03::read_file("$ROOTDIR/$PRONTUS_ID/$TMP_DIR/pags$VISTADIR/$plantilla\.$EXT");
        $NOM_PLT_EXITO = "$plantilla\.$EXT";
    } else {
        $TMP_EXITO = glib_fildir_03::read_file("$ROOTDIR/$PRONTUS_ID/$TMP_DIR/pags$VISTADIR/exito\.$EXT");
    }
    if ( $TMP_EXITO eq '') {
        lib_form::aborta(lib_lang::get_message('form_error_no_success_tpl'));
    }

    if ($RESERVED_DATA{'_pag_error'} ne '') {
        $plantilla = $RESERVED_DATA{'_pag_error'};
        $plantilla =~ s/[^\w\-]//g; # Solo caracteres alfanumericos y - y _.
        # Lee la plantilla del directorio de la vista.
        $TMP_ERROR = glib_fildir_03::read_file("$ROOTDIR/$PRONTUS_ID/$TMP_DIR/pags$VISTADIR/$plantilla\.$EXT");
        $NOM_PLT_ERROR = "$plantilla\.$EXT";
    } else {
        $TMP_ERROR = glib_fildir_03::read_file("$ROOTDIR/$PRONTUS_ID/$TMP_DIR/pags$VISTADIR/error\.$EXT");
    }

    if ( $TMP_ERROR eq '' ) {
        lib_form::aborta(lib_lang::get_message('form_error_no_error_tpl'));
    }

    if ($RESERVED_DATA{'_pag_estatico'} ne '' ) {
        $SALIDA_ESTATICA = 1;
    }

    # Busca, inicializa y elimina mensajes dentro de la plantilla.
    # Carga hash de mensajes (%MSGS).
    #     <!-- MSG xxx = xxx -->
    while ($TMP_ERROR =~ /<!--\s*MSG\s*(\w+)\s*=\s*(.+?)\s*-->/sg) {
        $MSGS{$1} = $2;
    }
    # Elimina mensajes de la plantilla.
    $TMP_ERROR =~ s/<!--\s*MSG\s*(\w+)\s*=\s*(.+?)\s*-->//sg;

    # inicializamos variables básicas de la lib_form desde las prontus_varglb.
    $lib_form::SERVER_SMTP = $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'SERVER_SMTP'};
    if ($lib_form::SERVER_SMTP eq '') {
        &salida($MSGS{'out_of_service'},$PRONTUS_VARS{'form_msg_error'.$VISTAVAR},$TMP_ERROR,1);
    }

    # Se valida el campo: _form_vista
    my $FORM_VISTA = $RESERVED_DATA{'_form_vista'};
    if($FORM_VISTA ne '' && $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'MULTIVISTA'}{$FORM_VISTA} != 1) {
        $MSGS{$FORM_VISTA} = glib_html_03::text2html($FORM_VISTA) unless($MSGS{$FORM_VISTA});
        &salida($MSGS{'wrong_vista'} . ' ' . $MSGS{$FORM_VISTA}, $PRONTUS_VARS{'form_msg_error'}, $TMP_ERROR,1);
    }

    # Chequea Captcha si es que es requerido
    if($PRONTUS_VARS{'chk_form_captcha_enable'} ne '') {
        if (!glib_cgi_05::param('g-recaptcha-response')) {
            my $captcha_type = 'form';
            my $captcha_input = $RESERVED_DATA{'_captcha_form'};
            $captcha_input = $RESERVED_DATA{'_captcha_text'} if($captcha_input eq '');

            my $captcha_img = $RESERVED_DATA{'_captcha_img'};
            my $captcha_code = $RESERVED_DATA{'_captcha_code'};
            my $msg_err_captcha = lib_captcha2::valida_captcha($captcha_input, $captcha_code, $captcha_type, $captcha_img);
            if ($msg_err_captcha ne '') {
                &salida($MSGS{'wrong_captcha'}, $PRONTUS_VARS{'form_msg_error'.$VISTAVAR}, $TMP_ERROR,1);
            }
        } else {
            # Se valida re-captcha para continuar
            $RECAPTCHA_RESPONSE = glib_cgi_05::param('g-recaptcha-response');
            my %form = (
                secret => $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'RECAPTCHA_SECRET_CODE'},
                response => $RECAPTCHA_RESPONSE
            );

            my $strjson = lib_prontus::post_http($prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'RECAPTCHA_API_URL'}, \%form);
            if ($strjson eq 'error') {
                lib_form::aborta(lib_lang::get_message('global_simple_error_retry'));
            }
            if ($strjson) {
                utf8::encode($strjson);

                my $hashtemp = eval { JSON::decode_json($strjson) };
                if ($@) {
                    lib_stdlog::print_err(lib_lang::get_interpolated_message('global_json_decode_error', $@, $strjson)."\n");
                    make_resp_and_exit(lib_lang::get_message('global_public_error_validating_grecaptcha'), 1);

                } elsif (defined $hashtemp->{'success'}) {
                    if(!$hashtemp->{'success'}){
                        my $message = '';
                        if (ref($hashtemp->{'error-codes'}) eq 'ARRAY') {
                            $message = join(', ', @{$hashtemp->{'error-codes'}});
                        } else {
                            lib_stdlog::print_err(lib_lang::get_interpolated_message('form_wrong_error_codes', ref($hashtemp->{'error-codes'})).".\n");
                        }
                        #lib_stdlog::print_err("Error al validar recaptcha google: [$message]\n");
                        &salida($MSGS{'wrong_google_captcha'}, $PRONTUS_VARS{'form_msg_error'.$VISTAVAR}, $TMP_ERROR,1);
                    }  else {
                        if (defined $hashtemp->{'score'}) { # recaptcha v3, debo validar score y action.
                            my $form_action = glib_cgi_05::param('g-recaptcha-action');
                            if ($hashtemp->{'score'} < $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'RECAPTCHA_V3_SCORE'}) {
                                #lib_stdlog::print_err("Invocación a Prontus Form validando recaptcha v3 con score [$hashtemp->{'score'}], menor a [$prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'RECAPTCHA_V3_SCORE'}]\n");
                                &salida($MSGS{'wrong_google_captcha'}, $PRONTUS_VARS{'form_msg_error'.$VISTAVAR}, $TMP_ERROR, 1);
                            }
                            if ($form_action ne $hashtemp->{'action'}) {
                                #lib_stdlog::print_err("Invocación a Prontus Form validando recaptcha v3 con action [$form_action], no coincide con la devuelta por google: [$hashtemp->{'action'}]\n");
                                &salida($MSGS{'wrong_google_captcha'}, $PRONTUS_VARS{'form_msg_error'.$VISTAVAR}, $TMP_ERROR, 1);
                            }
                        } else {
                            # en este caso es recaptcha v2 y la respuesta fue éxito, no hacemos nada más.
                            # lib_stdlog::print_err("Success usando recaptcha v2: [$strjson]\n");
                        }
                    }
                } else {
                    #lib_stdlog::print_err("Error al validar recaptcha google, no está definido el campo success en la respuesta.\n");
                    &salida($MSGS{'wrong_google_captcha'}, $PRONTUS_VARS{'form_msg_error'.$VISTAVAR}, $TMP_ERROR,1);
                }
            }
        }
    }

    # Chequea campos requeridos.
    if($PRONTUS_VARS{'chk_form_multivista_strict'}) {
        # lib_stdlog::print_err("chk_form_multivista_strict encontrado: [$PRONTUS_VARS{'chk_form_multivista_strict'}]\n");

        foreach my $key (keys %PRONTUS_VARS) {
            next unless(substr($key, 0, 17) eq 'chk_form_required');

            $nombre = substr($key, 18);
            #lib_stdlog::print_err("check encontrado: [$key]\n");
            #lib_stdlog::print_err("vistavar: [$VISTAVAR]\n");
            #lib_stdlog::print_err("nombre: [$nombre]\n");

            # Estamos en una vista, por lo tanto se valida sólo si el nombre termina en esa vista
            if($VISTAVAR) {

                if($nombre =~ /${VISTAVAR}$/) {
                    $MSGS{$nombre} = glib_html_03::text2html($nombre) unless($MSGS{$nombre});
                    if (glib_cgi_05::param($nombre) eq '') {
                        &salida($MSGS{'required_data'} .' '. $MSGS{$nombre}, $PRONTUS_VARS{'form_msg_error'.$VISTAVAR}, $TMP_ERROR, 1);
                    }
                }

            # Si no viene la vista no puede terminar en ninguna de las vistas
            } elsif($nombre =~ /_([^_]+?)$/) {
                my $posiblevista = $1;
                #lib_stdlog::print_err("posiblevista: $posiblevista\n");
                #lib_stdlog::print_err("validando: $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'MULTIVISTA'}{$posiblevista}\n");

                if(! $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'MULTIVISTA'}{$posiblevista}) {
                    $MSGS{$nombre} = glib_html_03::text2html($nombre) unless($MSGS{$nombre});
                    if (glib_cgi_05::param($nombre) eq '') {
                        &salida($MSGS{'required_data'} .' '. $MSGS{$nombre}, $PRONTUS_VARS{'form_msg_error'.$VISTAVAR}, $TMP_ERROR, 1);
                    }
                }
            }
        }

    } else {
        foreach my $key (keys %PRONTUS_VARS) {
            glib_str_03::correct_encode_utf8_string($key);
            if (substr($key, 0, 17) eq 'chk_form_required') {
                $nombre = substr($key, 18);
                $MSGS{$nombre} = glib_html_03::text2html($nombre) unless($MSGS{$nombre});
                if (glib_cgi_05::param($nombre) eq '') {
                    &salida($MSGS{'required_data'} .' '. $MSGS{$nombre}, $PRONTUS_VARS{'form_msg_error'.$VISTAVAR}, $TMP_ERROR, 1);
                }
            }
        }
    }

    # Valida los datos segun el tipo, el cual se deduce del nombre.
    foreach my $nombre (@DATOS) {
        $dato = glib_str_03::trim(glib_cgi_05::param($nombre)); # Elimina espacios para que no molesten.
        next if ($dato eq ''); # Solo valida datos ingresados.
        push @DATOS4IF, $nombre;
        my $nombre_lc = lc $nombre;
        if (substr($nombre_lc, 0, 3) eq 'rut') { # Valida RUTs.
            if (! lib_validator::chequea_rut($dato)) {
                my $nombre_dato = $MSGS{$nombre};
                $nombre_dato = glib_html_03::text2html($nombre) unless($nombre_dato);
                &salida($MSGS{'wrong_data'} .' '. $nombre_dato, $PRONTUS_VARS{'form_msg_error'.$VISTAVAR}, $TMP_ERROR, 1);
            }
        }
        if (($nombre_lc =~ /^tel[eé]fono/) || ($nombre_lc =~ /^fono/)
                || ($nombre_lc =~ /^phone/) || ($nombre_lc =~ /^celular/)
                || ($nombre_lc =~ /^fax/) || ($nombre_lc =~ /^cellphone/)) { # Valida telefonos.
            if (! lib_validator::chequea_telefono($dato)) {
                my $nombre_dato = $MSGS{$nombre};
                $nombre_dato = glib_html_03::text2html($nombre) unless($nombre_dato);
                &salida($MSGS{'wrong_data'} .' '. $nombre_dato, $PRONTUS_VARS{'form_msg_error'.$VISTAVAR}, $TMP_ERROR, 1);
            }
        }
        if (($nombre_lc =~ /^email/) || ($nombre_lc =~ /^e-mail/)) { # Valida emails.
            if (! lib_validator::chequea_email($dato)) {
                my $nombre_dato = $MSGS{$nombre};
                $nombre_dato = glib_html_03::text2html($nombre) unless($nombre_dato);
                &salida($MSGS{'wrong_data'} .' '. $nombre_dato, $PRONTUS_VARS{'form_msg_error'.$VISTAVAR}, $TMP_ERROR, 1);
            }
        }

        # Aprovecha de validar las extensiones de los archivos que se esten intentando subir.
        my $upload_filename = glib_cgi_05::real_paths($nombre);
        if (defined($upload_filename) && $upload_filename ne '') { # Es un archivo.
            if ($upload_filename !~ /(\.pdf|\.doc|\.docx|\.rtf|\.xls|\.xlsx|\.csv|\.zip|\.rar|\.jpg|\.jpeg|\.gif|\.png|\.bmp|\.txt|\.ppt|\.pptx|\.swf)$/i) {
                &salida(lib_lang::get_message('form_error_wrong_upload_file'), $PRONTUS_VARS{'form_msg_error'.$VISTAVAR}, $TMP_ERROR, 1);
            }
        }
    }
}

# Inicializa las variables de invocacion.
sub parse_request_data {
    my @list = ('_admin', '_captcha_code', '_captcha_form', '_captcha_img', '_captcha_text', '_file', '_form_vista', '_pag_error', '_pag_estatico', '_pag_exito', '_prontus_id', '_ts');

    # busca campos necesarios en minusculas y mayusculas, sino los asigna vacíos
    foreach my $key (@list) {
        my $value = glib_cgi_05::param($key);
        if (defined($value)) {
            $RESERVED_DATA{$key} = $value;
        } else {
            $value = glib_cgi_05::param(uc $key);
            if (defined($value)) {
                $RESERVED_DATA{$key} = $value;
            } else {
                $RESERVED_DATA{$key} = '';
            }
        }
    }

    $TS = $RESERVED_DATA{'_ts'}; # Timestamp.
    if ($TS !~ /^\d{14}$/) {
        lib_form::aborta(lib_lang::get_message('form_error_wrong_form'));
    }
    $FECHA = substr($TS,0,8);
    $PRONTUS_ID = $RESERVED_DATA{'_prontus_id'}; # Nombre del Prontus.
    if ($PRONTUS_ID eq '') {
        lib_form::aborta(lib_lang::get_message('global_invalid_prontus_id'));
    }

    if(! lib_utils::valida_prontus($PRONTUS_ID)) {
        lib_form::aborta(lib_lang::get_message('global_invalid_prontus_id'));
    }

    # Busca y lee las variables Prontus del articulo.
    # Lee las variables Prontus del articulo a partir de su xml.
    # Por simplicidad lo carga todo en el hash %PRONTUS_VARS.
    my $path_final_xml = "$ROOTDIR/$PRONTUS_ID/site/artic/$FECHA/xml/$TS\.xml";
    my $xml = lib_prontus::get_xml_data($path_final_xml);
    %PRONTUS_VARS = lib_prontus::getCamposXml($xml);
}

# Finaliza con un mensaje de exito o error entregado utilizando la plantilla
# de exito o error.
# Los parametros son: mensaje, plantilla de mensaje y plantilla de pagina.
sub salida {
    my ($msg,$string_error,$plantilla,$hay_error) = @_;

    # Define directorio de las respuestas y la identificacion de esta.
    # Solicitaron tener las respuestas en un dir separado para cada form, con el titular "slugificado".
    my $titular = $PRONTUS_VARS{'_txt_titular'};
    $titular = lib_prontus::ajusta_titular_friendly($titular);
    $ANSWERS_DIR = "/$PRONTUS_ID/$CACHE_DIR/$titular/exito";
    if ($hay_error) {
        # Para el error_log
        lib_stdlog::print_err("[prontus_form] Error: $msg\n");
        $ANSWERS_DIR = "/$PRONTUS_ID/$CACHE_DIR/$titular/error";
    }

    # Verifica que existe el directorios de cache y los crea si no es asi.
    if (! (-d "$ROOTDIR/$ANSWERS_DIR") ) {
        if (glib_fildir_03::check_dir("$ROOTDIR$ANSWERS_DIR") == 0) {
            lib_form::aborta(lib_lang::get_interpolated_message('form_error_answers_dir', $ANSWERS_DIR));
        }
    }

    $ANSWERID = $PRONTUS_ID . $TS . time . $$ . rand(1000000000);

    # Procesa IFs y NIFs.
    $plantilla = &procesaIFs($plantilla,2);
    $string_error = &procesaIFs($string_error,1);
    # Parsea los datos dentro de la plantilla y dentro del mensaje de exito.
    foreach my $key (@DATOS) {
        my $valor = glib_html_03::text2html(glib_str_03::trim(glib_cgi_05::param($key)));
        $valor =~ s/%/&#37;/g;
        $plantilla =~ s/%%\Q$key\E%%/$valor/sg;
        $string_error =~ s/%\Q$key\E%/$valor/sg;
        $key = lc $key;
        $plantilla =~ s/%%\Q$key\E%%/$valor/sg;
        $string_error =~ s/%\Q$key\E%/$valor/sg;
    }
    # Inserta mensaje de error.
    $string_error =~ s/%err%/$msg/g;
    # Elimina tags no parseados en el mensaje.
    $string_error =~ s/%\w+%//sg;

    # Sustituye mensaje en la plantilla.
    $plantilla =~ s/%%msg%%/$string_error/g;
    $plantilla =~ s/%%_referer%%/$ENV{'HTTP_REFERER'};/g;
    $plantilla =~ s/%%_answerpage%%/$ANSWERS_DIR\/$ANSWERID\.$EXT/g;
    $plantilla =~ s/%_pf_(\w+\(.*?\))%/%%_pf_$1%%/sg;
    $plantilla = lib_prontus::parser_custom_function($plantilla);

    # Elimina tags no parseados en la plantilla.
    $plantilla =~ s/%%\w+%%//sg;
    $plantilla =~ s/%\w+%//sg;

    # Salida estatica y existe plantilla exito estatica
    if ($SALIDA_ESTATICA && $NOM_PLT_EXITO && !$hay_error) {
        # Escribe el archivo de respuesta.
        my $archivo = "$ROOTDIR/$ANSWERS_DIR/$NOM_PLT_EXITO";
        glib_fildir_03::write_file($archivo, $plantilla);

        print "Status: 303 See Other\n";
        if($ENV{'QUERY_STRING'}){
            print "Location: $ANSWERS_DIR/$NOM_PLT_EXITO?$ENV{'QUERY_STRING'}\n\n";
        }else{
            print "Location: $ANSWERS_DIR/$NOM_PLT_EXITO\n\n";
        }
        exit;
    }

    # Salida estatica y existe plantilla exito estatica
    if ($SALIDA_ESTATICA && $NOM_PLT_ERROR && $hay_error) {
        # Escribe el archivo de respuesta.
        my $archivo = "$ROOTDIR/$ANSWERS_DIR/$NOM_PLT_ERROR";
        glib_fildir_03::write_file($archivo, $plantilla);

        print "Status: 303 See Other\n";
        if($ENV{'QUERY_STRING'}){
            print "Location: $ANSWERS_DIR/$NOM_PLT_ERROR?$ENV{'QUERY_STRING'}\n\n";
        }else{
            print "Location: $ANSWERS_DIR/$NOM_PLT_ERROR\n\n";
        }
        exit;
    }

    # Escribe el archivo de respuesta.
    my $archivo = "$ROOTDIR/$ANSWERS_DIR/$ANSWERID\.$EXT";
    glib_fildir_03::write_file($archivo, $plantilla);

    # Redirige al visitante hacia la pagina de respuesta.
    print "Status: 303 See Other\n";
    if($ENV{'QUERY_STRING'}){
        print "Location: $ANSWERS_DIR/$ANSWERID\.$EXT?$ENV{'QUERY_STRING'}\n\n";
    }else{
        print "Location: $ANSWERS_DIR/$ANSWERID\.$EXT\n\n";
    }

    # Limpia el directorio de archivos temporales.
    glib_fildir_03::garbage_collection("$ROOTDIR$ANSWERS_DIR");

    exit;
}

# Procesa los IFs y los NIFs en el string pasado como parametro, de acuerdo
# a si las variables estan o no presentes en @DATOS.
sub procesaIFs {
    my($buffer) = $_[0];
    my($numpers) = $_[1];
    my($regexp1,$regexp2,$pre,$post,$cuerpo,$tipo,$variable,$esta,$nombre,$valor);
    if ($numpers == 1) {
        $regexp1 = '%(n?if)\(([\w\=]+)\)%(.+?)%\/n?if%';
        $regexp2 = '%(n?if)\(([\w\=]+)\)%';
    } else {
        $regexp1 = '%%(n?if)\(([\w\=]+)\)%%(.+?)%%\/n?if%%';
        $regexp2 = '%%(n?if)\(([\w\=]+)\)%%';
    }
    # Busca el inicio y el fin de un IF o NIF.
    while ($buffer =~ /$regexp1/sg) {
        $pre = $`;
        $post = $';
        $tipo = lc $1;
        $variable = $2;
        $cuerpo = $3;
        # Avanza hasta el ultimo nivel de inicio de IF para permitir anidamiento.
        while($cuerpo =~ /$regexp2/sg) {
            $cuerpo = $';
            $pre .= "%$tipo($variable)%" . $`;
            $tipo = $1;
            $variable = $2;
        }
        # Determina y ejecuta el tipo de chequeo sobre la variable.
        if ($variable =~ /=/) {
            ($nombre,$valor) = split(/=/,$variable);
            if (glib_cgi_05::param($nombre) eq $valor) {
                $esta = 1;
            } else {
                $esta = 0;
            }
        } else {
            if (grep {$_ eq $variable} @DATOS4IF) {
                $esta = 1;
            } else {
                $esta = 0;
            }
        }
        # Reconstruye buffer de acuerdo a la logica.
        if ($tipo eq 'if') {
            if ($esta) {
                $buffer = $pre . $cuerpo . $post;
            } else {
                $buffer = $pre . $post;
            }
        } else {
            if ($esta) {
                $buffer = $pre . $post;
            } else {
                $buffer = $pre . $cuerpo . $post;
            }
        }
    }
    return $buffer;
}
