#!/usr/bin/perl

# ---------------------------------------------------------------
# Prontus CMS
# http://www.prontus.cl
# by Altavoz.net
#
# licensed under LGPL license.
# http://www.prontus.cl/license.html
# ---------------------------------------------------------------

# -------------------------------COMENTARIO GLOBAL---------------
# ---------------------------------------------------------------
# PROPOSITO .
# -----------
# Este script corre por consola, y tiene tres modos de funcionamiento:
# - Indexación masiva: si se le pasa el parámetro "full", indexa todos los documentos que puede.
#   Busca en la base de datos los artículos con alta, y los procesa
#   en batches del siguiente modo
#   - genera una salida NDJSON por batch
#   - gatilla un proceso curl por batch
# - Indexación de un artículo: si se le pasa como parámetro un ts, indexa ese documento (crea o actualiza el índice)
# - Eliminación de un artículo del índice: si se le pasa como parámetros un ts y la string DELETE, borra el documento del índice de búsqueda.
#
# Configuración:
#   Este script indexa contenidos basándose en
# - el archivo de configuración [webroot]/[prontus]/cpan/[prontus]-elasticsearch.cfg
#   Ese archivo debe existir, y su contenido debería ser al menos este:
# -----------
#   ELASTICSEARCH_FIDS = "fid_foto.html fid_noticia.html fid_video.html" # fids a indexar separados por espacios
#   ELASTICSEARCH_HOST = "10.0.0.20" # ip o hostname del servidor donde está elasticsearch
#   ELASTICSEARCH_HOST_PROTOCOL = "http" # protocolo del server elasticsearch.
#   ELASTICSEARCH_PORT = "9200" # puerto HTTP del servidor elasticsearch
#   ELASTICSEARCH_INDEX = "prontus_next" # nombre del index a crear. Aquí quedarán indexados los documentos de este prontus.
#   ELASTICSEARCH_TEXTVARS = "_txt_bajada:3 vtxt_cuerpo:2..." [nombre del campo]:[factor de boost].
#   ELASTICSEARCH_TITLEVAR = "_txt_titular" [por defecto el factor de boost es 10]
#   ELASTICSEARCH_METADATA[1..20] = ... # otros campos específicos. Acá habría que distinguir el tipo de campo.
# ----------
# - la configuración más abajo. Para adaptarlo a cada sitio en particular, modificar aquí al menos la configuración
#   $NDJSON_TEMPLATE y luego la sub procesar_sql agregando el código necesario para extraer cada campo del xml del artículo.
# Para usar este sistema, copiar este script a uno sin el nombre "-template" y ejecutarlo por consola:
#   Indexación masiva:
#       perl [path]/cgi-cpn/prontus_indexar_elasticsearch.pl [prontus] full # borra el indice antes de reindexar
#   Indexación específica:
#       perl [path]/cgi-cpn/prontus_indexar_elasticsearch.pl [prontus] [ts]
#   Borrado de índice de un artículo:
#       perl [path]/cgi-cpn/prontus_indexar_elasticsearch.pl [prontus] [ts] DELETE
# ---------------------------------------------------------------
use utf8;
use strict;

BEGIN {
    use FindBin qw($Bin);
    my $pathLibsProntus = $Bin;
    unshift(@INC,$pathLibsProntus);
    $pathLibsProntus =~ s/\/[^\/]+$//;
    unshift(@INC,$pathLibsProntus);
}

##libs##
use lib_stdlog;
lib_stdlog::set_stdlog($0, 51200);
use prontus_varglb;
prontus_varglb::init();

use Artic;
use DBI;
use Encode;
use File::Copy;
use HTTP::Request;
use JSON;
use LWP::ConnCache;
use LWP::UserAgent;
use POSIX;
use glib_dbi_03;
use glib_fildir_03;
use glib_hrfec_03;
use glib_str_03;
use lib_bd;
use lib_lang;
use lib_logproc;
use lib_prontus;
use lib_utils;
##/libs##

# variables globales de config por defecto.
# Si no se encuentran alternativas en el archivo de config [prontus]-elasticsearch.cfg, se usarán estas.
my $INVOCACION; # web o cli
my $ELASTICSEARCH_HOST_PROTOCOL = 'http';
my $ELASTICSEARCH_HOST = "localhost";
my $ELASTICSEARCH_PORT = "9200";
my $ELASTICSEARCH_INDEX = 'index_prontus';
my $ELASTICSEARCH_BATCH_CHUNK = 50; # procesamos 50 articulos por vez. En pruebas, CURL se cae si se le pasan strings demasiado largas. Reducir este valor o cambiar la cantidad/peso de los campos a indexar si sucede eso.
my %CAMPOS_PREDEFINIDOS; # Se rellenan con los campos predefinidos para comparar contra ELASTICSEARCH_SEARCH_META, para evitar errores al crear los mappings.

# Variables globales para configurar la indexación. Modificar acá.

# Configuraciones globales del index, como mappings, etc.
# Deberíamos crear el índice, luego crear los mapeos con JSON_TEMPLATE_BASE, luego indexar.
# Este doble indexado se hace para mostrar correctamente resultados con y sin acentos en la misma palabra. https://stackoverflow.com/questions/38452791/elasticsearch-insensitive-search-accents
my $SETTINGS = '{
    "settings": {
        "analysis": {
            "analyzer": {
                "folding": {
                    "tokenizer": "standard",
                    "filter": [
                        "lowercase",
                        "asciifolding",
                        "spanish_stop"
                    ]
                }
            },
            "filter": {
                "spanish_stop": {
                    "type": "stop",
                    "stopwords": "_spanish_"
                }
            }
        }
    }
}';

my $JSON_TEMPLATE_BASE = '
{
    "properties": {
        "_txt_titular": {
            "type": "text",
            "analyzer": "spanish",
            "fields": {
                "std":{
                    "type": "text",
                    "analyzer": "folding"
                }
            }
        },
        "_txt_bajada": {
            "type": "text",
            "analyzer": "spanish",
            "fields": {
                "std":{
                    "type": "text",
                    "analyzer": "folding"
                }
            }
        },
        "_fechap": {
            "type": "long"
        },
        "_priorizar": {
            "type": "boolean"
        }
        %%_properties%%
    }
}';

# JSON de base para campos de texto indexable y analizado en un idioma (por ahora español).
# Notar que se guardan dos campos: uno analizado en español, otro con analizador standard para
# matches más amplios.
my $JSON_TEXT_TEMPLATE = '    "%%nombre_campo%%": {
        "type": "text",
        "index": true,
        "analyzer": "spanish",
        "fields": {
            "std":{
                "type": "text",
                "analyzer": "folding"
            }
        }
    }';

# JSON de base para campos de texto no indexado.
my $JSON_KEYWORD_TEMPLATE = '
    "%%nombre_campo%%": {
        "type": "keyword",
        "index": false
    }
';

# JSON de base para campos de fecha.
my $JSON_DATE_TEMPLATE = '
    "%%nombre_campo%%": {
        "type": "date"
    }
';

# JSON de base para campos numéricos.
my $JSON_NUMBER_TEMPLATE = '
    "%%nombre_campo%%": {
        "type": "long"
    }
';

# Template para los campos a indexar. Este es un newline-delimited JSON, donde la primera línea indica el índice en que crear el documento, y la segunda contiene los datos a indexar.
# cada campo tiene el nombre de una marca Prontus.
# El contenido aquí son los campos de base, y pueden sumarse a éstos los campos en ELASTICSEARCH_TEXTVARS y ELASTICSEARCH_METADATA, por ejemplo.
my $NDJSON_TEMPLATE = '{"index": {"_index":"%%_index%%", "_id": "%%_ts%%"}}
{"_txt_titular": "%%_titular%%", "_txt_bajada": "%%_bajada%%", "_fileurl": "%%url%%", "_seccion1": "%%_seccion1%%", "_nom_seccion1": "%%_nom_seccion1%%", "_tema1": "%%_tema1%%", "_nom_tema1": "%%_nom_tema1%%", "_subtema1": "%%_subtema1%%", "_nom_subtema1": "%%_nom_subtema1%%", "_tags":[%%_tags%%], "_tagnames": [%%_tagnames%%], "_metadata_img": "%%_metadata_img%%", "_fechap":"%%_fechap%%", "_priorizar":"%%_priorizar%%" %%_otros_campos%%}
'; # NOTA: *no* va coma antes de %%_otros_campos%%, ya que esa variable puede estar vacía.

my $BD;

main: {
    lib_utils::setUp();
    my $prontus_id = $ARGV[0];
    $INVOCACION = $ARGV[1] || 'cli';

    my $ts = $ARGV[2] || '';
    my $action = $ARGV[3] || '';

    if ($prontus_id eq '') {
        print "Use: perl $0 <prontus_id>\n";
        exit;
    }

    # Carga variables de configuracion.
    my $msg_load = lib_utils::load_config_by_id($prontus_id);
    if ($msg_load ne '') {
        &gestionar_errores($msg_load, 1);
    }

    # Establece log file
    $lib_logproc::LOG_FILE = "$prontus_varglb::DIR_SERVER$prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'DIR_CPAN'}/procs/prontus_elasticsearch_reindex_log.html";
    $lib_logproc::JS_FILE  = "$prontus_varglb::DIR_SERVER$prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'DIR_CPAN'}/procs/elasticsearch_reindex.json";

    my $DIRPRONTUS = "$prontus_varglb::DIR_SERVER$prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'DIR_SITE'}$prontus_varglb::DIR_ARTIC";
    if (! -d $DIRPRONTUS) {
        &gestionar_errores(lib_lang::get_interpolated_message('elastic_index_invalid_dir', $DIRPRONTUS), 1);
    }

    # Conectar a BD
    my $msg_err_bd;
    ($BD, $msg_err_bd) = lib_bd::conectar_prontus_bd();
    if (! ref($BD)) {
        &gestionar_errores($msg_err_bd, 1);
    }

    $ELASTICSEARCH_HOST = $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'ELASTICSEARCH_HOST'};
    $ELASTICSEARCH_HOST_PROTOCOL = $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'ELASTICSEARCH_HOST_PROTOCOL'};
    $ELASTICSEARCH_PORT = $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'ELASTICSEARCH_PORT'};
    $ELASTICSEARCH_INDEX = $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'ELASTICSEARCH_INDEX'};

    # fids prontus a indexar.
    my @fids_validos = split(/[, ]/, $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'ELASTICSEARCH_FIDS'});
    my @tmp;
    # quitamos las eventuales extensiones de FID.
    foreach my $fid (@fids_validos) {
        if ($fid =~ /^(\w+)\.\w+$/) {
            push (@tmp, $1);
        } else {
            push (@tmp, $fid);
        }
    }
    @fids_validos = @tmp;

    # obtenemos:
    # - la string json que define cada campo en el mapping del index
    # - la string json que usaremos al indexar cada documento, con sus campos.
    # - un array de campos para procesar al leer los datos de cada documento.
    my ($index_properties, $string_campos, @campos_procesar) = &get_campos_procesar();
    #print STDERR "Properties: [$index_properties]\n\n";

    $JSON_TEMPLATE_BASE =~ s/%%_properties%%/$index_properties/g;
    $NDJSON_TEMPLATE =~ s/%%_otros_campos%%/$string_campos/g;

    my $ua = LWP::UserAgent->new();
    $ua->agent('Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692.99 Safari/537.36'); # default es "libwww-perl/#.##"
    my $conn_cache = LWP::ConnCache->new();
    $ua->conn_cache($conn_cache);
    my ($req, $method);
    my $header = ['Content-Type' => "application/json; charset=UTF-8"];
    my $url = "$ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT/$ELASTICSEARCH_INDEX";
    my $base_url = $url;
    # indexación masiva
    if ($ts eq 'full') { # procesamos todos los documentos válidos.
        my $command;
        my $start = 0;
        my $response; # para los resultados de los comandos.
        my $msg;
        if ($action =~ /^\d+$/) { # comenzar a procesar desde este item.
            $start = $action;
        } else { #recreamos el índice de cero.
            # borramos y recreamos antes el índice existente.
            $method = "DELETE";
            $req = HTTP::Request->new($method, $url, $header);
            $response = $ua->request($req);
            if (! $response->is_success && $response->code ne "404") {
                &gestionar_errores(lib_lang::get_interpolated_message('elastic_index_error_deleting', "1 $ELASTICSEARCH_INDEX", "$ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT: [$method] [$url] [".$response->status_line."]"), 1);
            }
            $response = JSON::decode_json($response->content);
            if ($response->{'error'} && $response->{'error'}{'type'} ne 'index_not_found_exception') {
                &gestionar_errores(lib_lang::get_interpolated_message('elastic_index_error_deleting', "2 $ELASTICSEARCH_INDEX", "$ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT: [$response->{'error'}{'type'}] [$response->{'error'}{'reason'}]"), 1);
            }

            &gestionar_errores(lib_lang::get_interpolated_message('elastic_index_previous_deleted', "$ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT/$ELASTICSEARCH_INDEX"), 0);
            $response = '';

            # creamos el índice
            $method = "PUT";
            $req = HTTP::Request->new($method, $url, $header, $SETTINGS);
            $response = $ua->request($req);
            if (!$response->is_success) {
                my $extra = '';
                if ($response->{'_content'}{'error'}{'root_cause'}) {
                    $extra = ", ".$response->{'_content'}{'error'}{'root_cause'}{'reason'};
                }

                &gestionar_errores(lib_lang::get_interpolated_message('elastic_index_error_config', "1 $ELASTICSEARCH_INDEX", "$ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT: [".$response->status_line."$extra]"), 1);
            }
            $response = JSON::decode_json($response->content);
            if ($response->{'error'}) {
                &gestionar_errores(lib_lang::get_interpolated_message('elastic_index_error_create', $ELASTICSEARCH_INDEX, "$ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT: [$response->{'error'}{'type'}] [$response->{'error'}{'reason'}]"), 1);
            }
            &gestionar_mensajes("Indice generado en $ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT/$ELASTICSEARCH_INDEX", 0);
            $response = '';

            # configuramos el índice con properties y mappings
            if ($JSON_TEMPLATE_BASE ne '') {
                # creamos el índice
                $method = "POST";
                $url = "$base_url/_mapping";
                $req = HTTP::Request->new($method, $url, $header, $JSON_TEMPLATE_BASE);
                $response = $ua->request($req);
                if (!$response->is_success) {
                    &gestionar_errores(lib_lang::get_interpolated_message('elastic_index_error_config', "2 $ELASTICSEARCH_INDEX", "$ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT: [".$response->status_line."]"), 1);
                }
                $response = JSON::decode_json($response->content);
                if ($response->{'error'}) {
                    &gestionar_errores(lib_lang::get_interpolated_message('elastic_index_error_config', "3 $ELASTICSEARCH_INDEX", "$ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT: [$response->{'error'}{'type'}] [$response->{'error'}{'reason'}]"), 1);
                }
                $response = '';

                &gestionar_mensajes(lib_lang::get_interpolated_message('elastic_index_mapping', "$ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT/$ELASTICSEARCH_INDEX"), 0);
            } else {
                &gestionar_mensajes(lib_lang::get_message('elastic_index_undefined_template'), 0);
            }
        }

        my $cantidad_articulos_inicial = &contar_articulos(\@fids_validos);
        $cantidad_articulos_inicial -= $start;
        &gestionar_mensajes(lib_lang::get_interpolated_message('elastic_index_count', $cantidad_articulos_inicial), 0);

        # procesamos los artículos en bloques
        my $template = "";
        my $string_articulos_a_indexar = '';
        $ELASTICSEARCH_BATCH_CHUNK = $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'ELASTICSEARCH_BATCH_CHUNK'};
        my $cantidad_articulos = $cantidad_articulos_inicial;
        my ($start_time, $end_time, $diff);
        while ($cantidad_articulos > 0) {
            $string_articulos_a_indexar = &generar_json_articulos(\@fids_validos, $start, $ELASTICSEARCH_BATCH_CHUNK, '', \@campos_procesar);
            $start += $ELASTICSEARCH_BATCH_CHUNK;
            &indexar_articulos($string_articulos_a_indexar, $ua);

            $cantidad_articulos -= $ELASTICSEARCH_BATCH_CHUNK;
            if ($cantidad_articulos < 0) {
                $cantidad_articulos = 0;
            }

            &gestionar_mensajes(lib_lang::get_interpolated_message('elastic_index_count', ($cantidad_articulos_inicial - $cantidad_articulos), $cantidad_articulos), 0);
        }
        &gestionar_mensajes(lib_lang::get_message('elastic_index_finish'), 1);
    } elsif ($ts =~ /^\d{4}$/) { # procesamos un año.
        my $anio = $ts;
        my $cantidad_articulos_inicial = &contar_articulos(\@fids_validos, $anio);
        &gestionar_mensajes(lib_lang::get_interpolated_message('elastic_index_count', $cantidad_articulos_inicial), 0);

        # procesamos los artículos en bloques
        my $start = 0;
        my $template = "";
        my $string_articulos_a_indexar = '';
        $ELASTICSEARCH_BATCH_CHUNK = $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'ELASTICSEARCH_BATCH_CHUNK'};
        my $cantidad_articulos = $cantidad_articulos_inicial;
        while ($cantidad_articulos > 0) {
            $string_articulos_a_indexar = &generar_json_articulos(\@fids_validos, $start, $ELASTICSEARCH_BATCH_CHUNK, $anio, \@campos_procesar);
            $start += $ELASTICSEARCH_BATCH_CHUNK;
            &indexar_articulos($string_articulos_a_indexar, $ua);

            &gestionar_mensajes(lib_lang::get_interpolated_message('elastic_index_count', ($cantidad_articulos_inicial - $cantidad_articulos), $cantidad_articulos), 0);

            $cantidad_articulos -= $ELASTICSEARCH_BATCH_CHUNK;
        }
        &gestionar_mensajes(lib_lang::get_message('elastic_index_finish'), 1);
    } elsif ($ts =~ /^\d{14}$/) { # procesamos un documento específico
        if ($action eq 'DELETE') { # eliminamos el documento
            my $command = "curl -XDELETE $ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT/$ELASTICSEARCH_INDEX/_doc/$ts";
            my $response = `$command`;
            my $msg;
            if ($?) {
                $msg = lib_lang::get_interpolated_message('elastic_index_error_delete_artic',
                    "[$ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT/$ELASTICSEARCH_INDEX/_doc/$ts]: $?");
                &gestionar_errores($msg, 1);
            }
            $response = JSON::decode_json($response);
            if ($response->{'error'}) {
                $msg = lib_lang::get_interpolated_message('elastic_index_error_delete_artic',
                    "[$ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT/$ELASTICSEARCH_INDEX/_doc/$ts]: [$response->{'error'}{'type'}] [$response->{'error'}{'reason'}]");
                &gestionar_errores($msg, 1);
            }
            $response = '';
            #print "Eliminando artículo [$ts] del índice.\n";
        } else { # creamos o actualizamos el índice del artículo.
            my $articulo_a_indexar = &generar_json_articulo(\@fids_validos, $ts, \@campos_procesar);
            if ($articulo_a_indexar) {
                &gestionar_mensajes(lib_lang::get_interpolated_message('elastic_index_update_artic', $ts), 0);
                &indexar_articulos($articulo_a_indexar, $ua);
            }
        }
    } else {
        &gestionar_errores(lib_lang::get_message('elastic_index_howto_use'), 1);
    }
}

# devuelve el conteo de artículos de los tipos especificados con alta = 1 en bd.
sub contar_articulos {
    my @fids_validos = @{$_[0]}; # referencia a array.
    my $anio = $_[1];
    my @params;
    my $ts_now = glib_hrfec_03::get_dtime_pack4();
    push (@params, $ts_now);
    push (@params, $ts_now); # se evalúa dos veces

    my $filtros;
    if (scalar @fids_validos > 0) {
        push (@params, @fids_validos);
        my $param_str = ', ?' x scalar @fids_validos;
        $param_str = substr $param_str, 2;
        $filtros = " and art_tipoficha in ($param_str) ";
    } else {
        $filtros = "";
    }

    # si pasaron un año, indexamos solamente ese año.
    my ($fecha_inicio, $fecha_fin);
    if ($anio =~ /^\d{4}$/) {
        $fecha_inicio = "${anio}0101";
        $fecha_fin = "${anio}1231";
        push (@params, $fecha_inicio);
        push (@params, $fecha_fin);
        $filtros .= " and (art_fechap >= ? and art_fechap <= ? )";
    }

    my $art_table = lib_utils::get_table('art');
    my $sql = "select count(*) from $art_table where art_alta = 1 and art_indexar = 1 and (concat(art_fechap,art_horap) <= ? and concat(art_fechae, art_horae) >= ?) $filtros";

    my $conteo;
    my $salida = glib_dbi_03::ejecutar_sql_bind_param($BD, $sql, \@params, \($conteo));
    $salida->fetch;
    $salida->finish;
    return $conteo;
}

# Recibe un array de FIDs que se pueden indexar, y un TS de un artículo.
# devuelve un string JSON con los datos ingresados según template.
# devuelve '' si el artículo no corresponde a un fid indexable.
sub generar_json_articulo {
    my @fids_validos = @{$_[0]}; # referencia a array.
    my $ts = $_[1];
    my @campos_procesar = @{$_[2]};

    my @params;

    my $ts_now = glib_hrfec_03::get_dtime_pack4();
    push (@params, $ts_now);
    push (@params, $ts_now); # se evalúa dos veces

    if ($ts !~ /\d{14}/) {
        return '';
    } else {
        push (@params, $ts);
    }

    my $filtros;
    if (scalar @fids_validos > 0) {
        push (@params, @fids_validos);
        my $param_str = ', ?' x scalar @fids_validos;
        $param_str = substr $param_str, 2; # eliminamos la primera ', '
        $filtros = " and art_tipoficha in ($param_str) ";
    } else {
        $filtros = "";
    }

    my $art_table = lib_utils::get_table('art');
    my $sql = "select substr(art_id, 1, 8) as art_fechac, art_id, art_tipoficha, art_titu, art_baja, art_priorizar from $art_table where art_alta = 1 and art_indexar = 1 and (concat(art_fechap,art_horap) <= ? and concat(art_fechae, art_horae) >= ?) and art_id = ? $filtros order by art_id limit 1";
    #print STDERR "$sql\n";
    my $json_articulo = &procesar_sql($sql, \@params, $ts, \@campos_procesar);

    return $json_articulo;
}

# Recibe una lista de fids válidos para indexar, y los valores de inicio y límite de procesamiento para la query sql
# que buscará los artículos.
# devuelve un string JSON con los datos ingresados según template.
sub generar_json_articulos {
    my @fids_validos = @{$_[0]}; # referencia a array.
    my $start = $_[1];
    my $limit = $_[2];
    my $anio = $_[3];
    my @campos_procesar = @{$_[4]};

    my @params;

    my $ts_now = glib_hrfec_03::get_dtime_pack4();
    push (@params, $ts_now);
    push (@params, $ts_now); # se evalúa dos veces

    my $filtros;
    if (scalar @fids_validos > 0) {
        push (@params, @fids_validos);
        my $param_str = ', ?' x scalar @fids_validos;
        $param_str = substr $param_str, 2; # eliminamos la primera ', '
        $filtros = " and art_tipoficha in ($param_str) ";
    } else {
        $filtros = "";
    }

    # si pasaron un año, indexamos solamente ese año.
    my ($fecha_inicio, $fecha_fin);
    if ($anio =~ /^\d{4}$/) {
        $fecha_inicio = "${anio}0101";
        $fecha_fin = "${anio}1231";
        push (@params, $fecha_inicio);
        push (@params, $fecha_fin);
        $filtros .= " and (art_fechap >= ? and art_fechap <= ? )";
    }

    push (@params, $start);
    push (@params, $limit);
    my $art_table = lib_utils::get_table('art');
    my $sql = "select substr(art_id, 1, 8) as art_fechac, art_id, art_tipoficha, art_titu, art_baja, art_priorizar from $art_table where art_alta = 1 and art_indexar = 1 and (concat(art_fechap,art_horap) <= ? and concat(art_fechae, art_horae) >= ?) $filtros order by art_id limit ?, ?";
    my $json_articulos = &procesar_sql($sql, \@params, '', \@campos_procesar);

    return $json_articulos;
}

# Recibe una sentencia SQL con parámetros y procesa la query, armando una string json con los resultados.
# Si recibe un ts prontus, busca los datos de ese artículo solamente
# De otro modo procesa todos los artículos con alta.
# devuelve una string NDJSON.
sub procesar_sql {
    my $sql = $_[0];
    my $params = $_[1];
    my $ts = $_[2];
    my @campos_procesar = @{$_[3]};

    my ($fechac, $fid, $titular, $bajada, $vtxt_cuerpo, $es_priorizado);
    my $salida = glib_dbi_03::ejecutar_sql_bind_param($BD, $sql, $params, \($fechac, $ts, $fid, $titular, $bajada, $es_priorizado));
    my $json_articulos = '';
    my $ndjson_template = $NDJSON_TEMPLATE;
    my $buffer = '';
    my $path = '';
    my $xmlbuffer = '';
    my %hashprontus;
    my $friendly_url = '';
    while ($salida->fetch) {
        $titular = &procesar_campo($titular);
        $bajada = &procesar_campo($bajada);

        $buffer = $ndjson_template;
        $buffer =~ s/%%_index%%/$ELASTICSEARCH_INDEX/;
        $buffer =~ s/%%_ts%%/$ts/;
        $buffer =~ s/%%_titular%%/$titular/;
        $buffer =~ s/%%_bajada%%/$bajada/;

        if ($es_priorizado == 0) {
            $es_priorizado = 'false';
        } else {
            $es_priorizado = 'true';
        }
        $buffer =~ s/%%_priorizar%%/$es_priorizado/;

        $path = "$prontus_varglb::DIR_SERVER/$prontus_varglb::PRONTUS_ID/site/artic/$fechac/xml/$ts.xml";
        # instanciamos un objeto ARTIC para generar la URL al artículo.
        my $artic_obj;
        # se instancia el articulo para obtener los datos
        unless($artic_obj = Artic->new(
                    'prontus_id'         => $prontus_varglb::PRONTUS_ID,
                    'public_server_name' => $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'PUBLIC_SERVER_NAME'},
                    'cpan_server_name'   => $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'CPAN_SERVER_NAME'},
                    'document_root'      => $prontus_varglb::DIR_SERVER,
                    'ts'                 => $ts,
                    'no_check_dirs'      => 1)) {

            &gestionar_errores(lib_lang::get_interpolated_message('global_error_artic_init', $Artic::ERR), 1, 0);
            return;
        };

        my $friendly_url = $artic_obj->get_friendly_url();
        my %hashprontus = $artic_obj->get_xml_content();

        # campos predefinidos a indexar o agregar: secciones, tags, imagen para resultados, fechap.
        my $seccion1 = $hashprontus{'_seccion1'} || '';
        my $nom_seccion1 = $hashprontus{'_nom_seccion1'};
        my $tema1 = $hashprontus{'_tema1'} || '';
        my $nom_tema1 = $hashprontus{'_nom_tema1'};
        my $subtema1 = $hashprontus{'_subtema1'} || '';
        my $nom_subtema1 = $hashprontus{'_nom_subtema1'};
        my $metadata_img = $hashprontus{$prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{"ELASTICSEARCH_METADATA_IMG"}};
        my $fechap = $hashprontus{'_fechap'};

        # separamos los tags y tagnames y los volvemos a unir rodeados por comillas.
        my @tags = split(',', $hashprontus{'_tags'});
        my $tags = '"'.join('","', @tags).'"';
        my @tagnames = split(',', $hashprontus{'_tagnames'});
        my $tagnames = '"'.join('","', @tagnames).'"';

        $buffer =~ s/%%_seccion1%%/$seccion1/;
        $buffer =~ s/%%_nom_seccion1%%/$nom_seccion1/;
        $buffer =~ s/%%_tema1%%/$tema1/;
        $buffer =~ s/%%_nom_tema1%%/$nom_tema1/;
        $buffer =~ s/%%_subtema1%%/$subtema1/;
        $buffer =~ s/%%_nom_subtema1%%/$nom_subtema1/;
        $buffer =~ s/%%_tags%%/$tags/;
        $buffer =~ s/%%_tagnames%%/$tagnames/;
        $buffer =~ s/%%_metadata_img%%/$metadata_img/;
        $buffer =~ s/%%_fechap%%/$fechap/;
        $buffer =~ s/%%url%%/$friendly_url/;

        %CAMPOS_PREDEFINIDOS = {
            '_seccion1' => 1,
            '_nom_seccion1' => 1,
            '_tema1' => 1,
            '_nom_tema1' => 1,
            '_subtema1' => 1,
            '_nom_subtema1' => 1,
            '_fechap' => 1,
            '_tags' => 1,
            '_tagnames' => 1,
        };

        # procesamos los demás campos agregados:
        foreach my $nombre_campo (@campos_procesar) {
            next if ($CAMPOS_PREDEFINIDOS{$nombre_campo});
            my $campo = &procesar_campo($hashprontus{$nombre_campo});
            $campo = substr($campo, 0, $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'ELASTICSEARCH_MAXCARS'});
            $buffer =~ s/%%$nombre_campo%%/$campo/g;
        }
        $buffer =~ s/%%.*%%//g;
        $json_articulos .= $buffer;
    }
    $json_articulos .= "\n";
    return $json_articulos;
}

# Recibe una string NDJSON con los artículos a indexar y gatilla una llamada a curl para ejecutar esa acción.
# TODO: implementar API Perl.
sub indexar_articulos {
    my ($json_indexar, $ua) = @_;
    # el parámetro "refresh" actualiza el shard con el documento para que pueda encontrarse inmediatamente.
    my $base_url = "$ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT/$ELASTICSEARCH_INDEX";
    my $url = $base_url."/_bulk?refresh";
    my $header = ['Content-Type' => "application/json; charset=UTF-8"];
    my $method = "POST";

    my $req = HTTP::Request->new($method, $url, $header, Encode::encode_utf8($json_indexar));

    my $response = $ua->request($req);

    # errores en el comando curl.
    # usualmente cuando falla aquí se debe a que la string de argumentos pasada por línea de comandos a curl es demasiado grande. Hay que jugar entonces con la cantidad de documentos indexados (ELASTICSEARCH_BATCH_CHUNK) o con el máximo de caracteres por campo (ELASTICSEARCH_MAXCARS).
    if (! $response->is_success) {
        &gestionar_errores(lib_lang::get_interpolated_message('elastic_index_error_indexing', "$ELASTICSEARCH_HOST_PROTOCOL://$ELASTICSEARCH_HOST:$ELASTICSEARCH_PORT/$ELASTICSEARCH_INDEX: [".$response->status_line."]"), 1);
    }
    # revisamos el resultado para comprobar que indexó.
    $response = JSON::decode_json($response->content);
    # errores globales al indexar, e.g. hay un error de sintaxis en el json.
    if ($response->{'error'}) {
        &gestionar_errores(lib_lang::get_interpolated_message('elastic_index_error_index_fail', "$response->{'error'}{'type'}; $response->{'error'}{'reason'}"), 1);
    }
    # errores particulares al indexar, e.g. incluimos dos veces un campo.
    if ($response->{'errors'}) {
        &gestionar_errores(lib_lang::get_interpolated_message('elastic_index_error_index_fail_extended',
            $response->{'items'}[0]{'index'}{'_id'},
            $response->{'items'}[0]{'index'}{'error'}{'type'},
            $response->{'items'}[0]{'index'}{'error'}{'reason'},
            $response->{'items'}[0]{'index'}{'error'}{'caused_by'}{'reason'}), 1);
    }
}

# Recibe un campo string y lo procesa para agregarlo correctamente
# al ndjson.
sub procesar_campo {
    my $campo = shift;
    $campo =~ s/['"\\]//g;
    $campo = lib_prontus::get_minitext_value($campo);
    $campo = lib_prontus::get_jsontext_value($campo);
    glib_str_03::correct_encode_utf8_string($campo);
    if ($campo eq 'null') {
        $campo = '';
    }
    return $campo;
}

# procesa los campos a indexar (textvars, metadata) desde las variables de config Prontus.
sub get_campos_procesar {
    # hash de campos a indexar y sus factores de boost.
    my %campos_a_indexar = ('_txt_titular' => 6, '_txt_bajada' => 5); # por defecto indexamos el titular con factor boost = 6 y la bajada con boost 5.
    my $index_properties = '';
    my $string_campos = '';
    my @campos_procesar;

    foreach my $campo (split(/\s+/, $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'ELASTICSEARCH_TEXTVARS'})) {
        my $template = $JSON_TEXT_TEMPLATE;
        my ($nombre, $boost) = split(":", $campo);

        next if ($nombre eq '_txt_titular' || $nombre eq '_txt_bajada'); # descartamos el titular y bajada, se indexan por defecto.

        $boost = POSIX::floor($boost); # convertimos a entero.
        $boost = 1 if (!$boost || $boost < 1); # el factor de potencia debe ser al menos 1.
        $campos_a_indexar{$nombre} = $boost;

        $template =~ s/%%nombre_campo%%/$nombre/g;
        $index_properties .= ",$template";
        $string_campos .= ", \"$nombre\": \"%%$nombre%%\"";
        push (@campos_procesar, $nombre);
    }

    # procesamos campos predefinidos numéricos. Por ahora sin uso. Nota: el timestamp yyyymmdd de fecha de publicación se procesa como campo numérico predefinido.
    my %campos_num = ();
    foreach my $campo (keys %campos_num) {
        my $metadata_template = $JSON_NUMBER_TEMPLATE;
        $campos_a_indexar{$campo} = $campos_num{$campo};
        $metadata_template =~ s/%%nombre_campo%%/$campo/g;
        $index_properties .= ",$metadata_template";
        $string_campos .= ", \"$campo\": \"%%$campo%%\"";
        push (@campos_procesar, $campo);
    }

    # searchable metadata
    for (my $i = 1; $i <= 5; $i++) {
        my $metadata_template = '';
        if ($prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{"ELASTICSEARCH_SEARCH_META$i"} ne '') {
            my $nombre_metadata = $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{"ELASTICSEARCH_SEARCH_META$i"};
            my $tipo_metadata = $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{"ELASTICSEARCH_SEARCH_META${i}_TIPO"};
            if ($tipo_metadata eq 'date') {
                $metadata_template = $JSON_DATE_TEMPLATE;
            } elsif ($tipo_metadata eq 'float' || $tipo_metadata eq 'number') {
                $metadata_template = $JSON_NUMBER_TEMPLATE;
            } else {
                $metadata_template = $JSON_KEYWORD_TEMPLATE;
            }
            $campos_a_indexar{$prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{"ELASTICSEARCH_SEARCH_META$i"}} = $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{"ELASTICSEARCH_SEARCH_META${i}_TIPO"};
            $metadata_template =~ s/%%nombre_campo%%/$nombre_metadata/g;
            $index_properties .= ",$metadata_template";
            $string_campos .= ", \"$nombre_metadata\": \"%%$nombre_metadata%%\"";
            push (@campos_procesar, $nombre_metadata);
        }
    }

    # otros campos de metadata (no queryable)
    for (my $i = 1; $i <= 10; $i++) {
        my $metadata_template = '';
        if ($prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{"ELASTICSEARCH_METADATA$i"} ne '') {
            my $nombre_metadata = $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{"ELASTICSEARCH_METADATA$i"};
            my $tipo_metadata = 'keyword';
            $metadata_template = $JSON_KEYWORD_TEMPLATE;
            $campos_a_indexar{$prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{"ELASTICSEARCH_METADATA$i"}} = $tipo_metadata;
            $metadata_template =~ s/%%nombre_campo%%/$nombre_metadata/g;
            $index_properties .= ",$metadata_template";
            $string_campos .= ", \"$nombre_metadata\": \"%%$nombre_metadata%%\"";
            push (@campos_procesar, $nombre_metadata);
        }
    }

    return ($index_properties, $string_campos, @campos_procesar);
}

sub gestionar_errores {
    my $mensaje = $_[0];
    my $complete = $_[1];
    my $exit = defined($_[2]) ? $_[2] : 0;
    if ($INVOCACION eq 'web') {
        lib_logproc::add_to_log("[error]: $mensaje");
        lib_logproc::showProgress($mensaje, $complete);
    } else {
        print STDERR "[error]: $mensaje\n";
    }
    exit if $exit;
}

sub gestionar_mensajes {
    my ($mensaje, $complete) = @_;
    if ($INVOCACION eq 'web') {
        lib_logproc::add_to_log($mensaje);
        lib_logproc::showProgress($mensaje, $complete);
    } else {
        print "$mensaje\n";
    }
}
