#!/usr/bin/perl
#
# Handler de CGIs Perl.
# Las invoca a traves de la interfaz FCGI, manejando paralelismo (workers)
# mediante FCGI::ProcManager.
#
# Se ejecuta a nombre del usuario www-data, idealmente mediante un watchdog.
# Como parametro se le pasa la cantidad de workers a crear. Por defecto es 1.
#
# Ejemplo, para crear 10 workers:
#   /usr/bin/perl /var/www/prontus/web/cgi-cpn/fcgi_fpm/fcgi_fpm.pl 10
#
# Versiones:
# 1.0 2021-11-02 ALD - Primera version.
#
# TODO:
#   - Reciclar workers despues de una cantidad de invocaciones.

use strict;
BEGIN {
    use FindBin qw($Bin);
    my $cgi_path = $Bin;
    unshift(@INC, $cgi_path);
    # Retrocede 1 directorio para quedar en /cgi-cpn.
    $cgi_path =~ s/\/[^\/]+\/?$//;
    # Para dejar disponibles las librerias
    unshift(@INC, $cgi_path);

    # Agrega directorios de bibliotecas a @INC.
    opendir(my $DIR, $cgi_path);
    my @entries = readdir($DIR);
    closedir $DIR;

    foreach my $entry (@entries) {
        next if ($entry =~ /^[\._]/);
        next if ($entry eq 'prontus_error_log' or $entry eq 'prontus_temp');
        if (-d "$cgi_path/$entry") {
            unshift(@INC, "$cgi_path/$entry");
        }
    }

    if (-d "$cgi_path/prontus_plugins") {
        opendir(my $DIR, "$cgi_path/prontus_plugins");
        @entries = readdir($DIR);
        closedir $DIR;

        foreach my $entry (@entries) {
            next if ($entry =~ /^[\._]/);
            next if ($entry eq 'prontus_error_log' or $entry eq 'prontus_temp');
            if (-d "$cgi_path/prontus_plugins/$entry") {
                unshift(@INC, "$cgi_path/prontus_plugins/$entry");
            }
        }
    }

    # Cambia a cgi-bin
    $cgi_path =~ s/\/cgi-cpn$/\/cgi-bin/;

    if (-d $cgi_path) {
        # la agrega a @INC.
        unshift(@INC, $cgi_path);
        opendir(my $dh, $cgi_path);
        @entries = readdir($dh);
        closedir $dh;
        foreach my $entry (@entries) {
            next if ($entry =~ /^[\._]/);
            next if ($entry eq 'prontus_error_log' or $entry eq 'prontus_temp');
            if (-d "$cgi_path/$entry") {
                unshift(@INC, "$cgi_path/$entry");
            }
        }
    }
}

use POSIX qw(:signal_h);

use sigtrap 'handler' => \&signal_term, 'INT';
use sigtrap 'handler' => \&signal_term, 'TERM';
use sigtrap 'handler' => \&signal_term, 'KILL';
use sigtrap 'handler' => \&signal_usr1, 'USR1';
use sigtrap 'handler' => \&signal_usr2, 'USR2';

use FCGI;
use FCGI::ProcManager;

use lib_stdlog;
lib_stdlog::set_stdlog($0, 5120000);
use prontus_varglb;
prontus_varglb::init();

# librerias de prontus
use Artic;
use Crypt::Eksblowfish::Bcrypt;
use Crypt::JWT;
use Data::Dumper;
use DateTime; #lib_stats
use DBI;
use Digest::MD5;
use Digest::SHA;
use Email::MIME;
use Email::Sender::Simple;
use Email::Sender::Transport::SMTP;
use Email::Simple;
use Encode;
use Fcntl qw(:flock SEEK_END);
use File::Compare;
use File::Copy;
use File::Find;
use File::LibMagic;
use GD;
use GD::Image;
use Getopt::Long;
use glib_cgi_05;
use glib_dbi_03;
use glib_fildir_03;
use glib_hrfec_03;
use glib_html_03;
use glib_str_03;
use HTML::Entities;
use HTTP::Request;
use HTTP::Request::Common;
use HTTP::Response;
use Image::ExifTool; # soporte webp y rotacion
use IO::All;
use IO::Compress::Gzip qw(gzip $GzipError); # para sitemaps
use JSON;
use lib_api_seo;
use lib_api_snp;
use lib_artic;
use lib_bd;
use lib_captcha2;
use lib_cookies;
use lib_dam;
use lib_dropbox;
use lib_edit;
use lib_encodes;
use lib_fid;
use lib_form;
use lib_friendly;
use lib_helper;
use lib_ipcheck;
use lib_lang;
use lib_loading;
use lib_lock;
use lib_logproc;
use lib_mail;
use lib_mam;
use lib_maxrunning;
use lib_multiediting;
use lib_multitag;
use lib_oidc;
use lib_parser;
use lib_profile;
use lib_ports;
use lib_prontus;
use lib_quota;
use lib_robots;
use lib_search;
use lib_setbd;
use lib_sitemap;
use lib_stats;
use lib_tags;
use lib_tax;
use lib_thumb;
use lib_upload;
use lib_urlf5;
use lib_users;
use lib_utils;
use lib_validator;
use lib_waitlock;
use lib_xcoding;
use LWP::ConnCache;
use LWP::UserAgent;
use MIME::Base64;
use MIME::Parser;
use mime_types;
use Net::Amazon::S3; # requerido por lib S3
use Net::Amazon::S3::Client; # requerido por lib S3
use POSIX;
use prontus_auth;
use Session;
use Time::HiRes;
use Time::Local;
use Update;
use URI::Escape;
use UUID4::Tiny;
use wizard_lib;
use XML::Parser;

eval {require Archive::Zip;};
if ($@) {
    $lib_prontus::ZIP_AVAILABLE = 0;
} else {
    Archive::Zip->import();
    $lib_prontus::ZIP_AVAILABLE = 1;
}

my $DEBUG = 0;
my $PROC_MANAGER;
my $SOCKET;
my $REQUEST;
my $SINGLE_LOG;
# Hash que mantiene la referencia a los scripts ya compilados.
my %FCGI_PROCESSES;
my $FPM_FILE;
my $SIGSET;

main: {
    my %conf = %{lib_utils::getHash("$prontus_varglb::DIR_SERVER/$prontus_varglb::DIR_CGI_CPAN/fcgi_fpm/fcgi.cfg")};

    my $numprocs = $prontus_varglb::FCGI_CONFIG{'num_procs_prontus'};
    # Asigna valor de configuracion si existe
    if (defined($conf{'num_procs_prontus'})) {
        $numprocs = $conf{'num_procs_prontus'};
    }

    if (defined($conf{'debug'})) {
        $DEBUG = $conf{'debug'};
    }

    $prontus_varglb::FCGI_MODE = 1;

    my $front = $ARGV[0];
    if ($front && defined($conf{'num_procs_front'})) {
        $numprocs = $conf{'num_procs_front'};
    }

    # Construye el directorio del socket.
    my $sockdir = $prontus_varglb::DIR_SERVER;
    $sockdir =~ s|/web.*|/run|;

    my $socket_path = "$sockdir/fcgi_fpm.socket";
    if ($front) {
        $socket_path = "$sockdir/fcgi_fpm_front.socket";
    }

    $SOCKET = FCGI::OpenSocket($socket_path, 10 );
    $REQUEST = FCGI::Request( \*STDIN, \*STDOUT, \*STDERR, \%ENV, $SOCKET, FCGI::FAIL_ACCEPT_ON_INTR );

    $FPM_FILE = $0;

    $SINGLE_LOG = $prontus_varglb::FCGI_CONFIG{'single_log'};
    # Asigna valor de configuracion si existe
    if (defined($conf{'single_log'})) {
        $SINGLE_LOG = $conf{'single_log'};
    }

    my $procName = $prontus_varglb::FCGI_CONFIG{'fcgi_proc_name'};
    my $workerName = $prontus_varglb::FCGI_CONFIG{'fcgi_worker_name'};
    # Asigna valor de configuracion si existe
    if (defined($conf{'procs_id'})) {
        $procName = $prontus_varglb::FCGI_CONFIG{'fcgi_proc_name'} . '-' .$conf{'procs_id'};
        $workerName = $prontus_varglb::FCGI_CONFIG{'fcgi_worker_name'} . '-' .$conf{'procs_id'};
    }

    if ($front) {
        $procName = $procName . "-front";
        $workerName = $workerName . "-front";
    }

    lib_stdlog::print_err("[" . glib_hrfec_03::get_dtime_pack4() . "] FCGI_FPM Master Start [$procName]\n");

    # pre-cargamos los idiomas disponibles
    my %languages = %{lib_lang::get_available_languages()};
    # procesamos la lista de idiomas
    foreach my $lang_key (keys %languages) {
        lib_lang::load_translation_table($lang_key);
    }

    $PROC_MANAGER = FCGI::ProcManager->new({
        n_processes => $numprocs,
        pm_title => $procName
    });

    $PROC_MANAGER->pm_manage();
    $PROC_MANAGER->pm_change_process_name($workerName);
    lib_stdlog::print_err("[" . glib_hrfec_03::get_dtime_pack4() . "] FCGI_FPM Worker Start [$$][$procName]\n");

    $SIGSET = POSIX::SigSet->new(SIGUSR1);
    $SIGSET->addset(SIGUSR2);
    &handle_requests();
}

sub handle_requests {
    lib_stdlog::print_err("FCGI_FPM START\n") if ($DEBUG);
    my $cc = 0;
    while ($REQUEST->Accept() >= 0) {
        $PROC_MANAGER->pm_pre_dispatch();
        sigprocmask(SIG_BLOCK, $SIGSET);

        my $script_file = $ENV{'SCRIPT_FILENAME'};

        if ((-s $script_file) && (-r $script_file)) {
            if ($SINGLE_LOG) {
                lib_stdlog::set_stdlog($FPM_FILE, 5120000);
            } else{
                lib_stdlog::set_stdlog($script_file, 51200);
            }

            $cc++;
            # Si no existe la referencia, compila el script.
            if (! $FCGI_PROCESSES{$script_file}) {
                lib_stdlog::print_err("FCGI_FPM pre-compile [$cc] [$script_file]\n") if ($DEBUG);
                my $script_content = glib_fildir_03::read_file($script_file);
                $script_content = &clean_script($script_content);
                # lib_stdlog::print_err("\n$script_content\n") if ($DEBUG);
                $FCGI_PROCESSES{$script_file} = eval "sub { $script_content }";
                lib_stdlog::print_err("FCGI_FPM compile-end [$cc] [$script_file][$@]\n") if ($DEBUG);
            }

            # Pasa el id del proceso como variable de ambiente, para identificar el fork.
            $ENV{'PROCESS_NUMBER'} = $$;
            # Ejecuta el script.
            lib_stdlog::print_err("[$$] FCGI_FPM serv [$cc] [$script_file][$@]\n") if ($DEBUG);
            $FCGI_PROCESSES{$script_file}->();

        }else{
            print "Content-type: text/plain\n\nNo such file or no exec permissions on it.\n";
        }

        EXIT_OVERRIDE: {
            lib_stdlog::print_err("FCGI_FPM exit $cc [$script_file][$@]\n") if ($DEBUG);
        }

        close $lib_stdlog::STDERR;
        $lib_stdlog::STDERR = undef;

        sigprocmask(SIG_UNBLOCK, $SIGSET);
        $PROC_MANAGER->pm_post_dispatch();
    }

    FCGI::CloseSocket($SOCKET);
}

# Elimina codigo de scripts para poder ejecutarlos sin problemas
sub clean_script {
    my $script = $_[0];
    # Quita bloque begin
    $script =~ s/BEGIN\s*\{[^}]+\}//s;
    # Quita uses de librerias
    $script =~ s|##libs##.+?##/libs##||s;

    $script =~ s|exit;|goto EXIT_OVERRIDE;|s;

    return $script;
}

sub signal_usr1 {
    lib_stdlog::print_err("[$$] Reload config\n");
    # limpia variables de configuracion para que sean recargadas
    lib_prontus::reset_prontus_config();
    # desbloquea señal para recibirla nuevamente
    sigprocmask(SIG_UNBLOCK, $SIGSET);
    # reinicia proceso
    &handle_requests();
}

sub signal_usr2 {
    lib_stdlog::print_err("[$$] Reset cache vars\n");
    # limpia variables de cache de prontus
    lib_prontus::reset_prontus_vars();
    # desbloquea señal para recibirla nuevamente
    sigprocmask(SIG_UNBLOCK, $SIGSET);
    # reinicia proceso
    &handle_requests();
}

# funcion para capturar las señales INT y TERM y logearlas
sub signal_term {
    lib_stdlog::print_err("STDERR: [$$] Ended by signal", @_ , "\n");
    $PROC_MANAGER->pm_exit();
    exit(0);
}
