#!/usr/bin/perl

# Handler para convertir URLs friendly arbitrarias a referencias a archivos
# fisicos, usando el internal_redirect para enviar el resultado al cliente.

# Version Apache:
# 1.0 2019-06-27 ALD - Primera version
# 1.1 2020-03-03 ALD - Implementa redirect si friendly trae un trailing slash
#                      (manejado dentro de lib_friendly.pm)
# 1.2 2020-03-03 ALD - Implementa respuesta 7 en lugar de lo anterior.
# 1.3 2020-06-04 ALD - Incluye el query string en el internal redirect.

# Version FCGI:
# 01 2021-10-19 ALD - Primera version.

# TODO:
# - Leer la cookie "vista" para determinar la vista.

use strict;

BEGIN {
    use FindBin qw($Bin);
    my $pathLibsProntus = $Bin;
    unshift(@INC,$pathLibsProntus);
    $pathLibsProntus =~ s/\/[^\/]+\/?$//;
    # Para dejar disponibles las librerias de prontus
    unshift(@INC,$pathLibsProntus);
}

use POSIX qw(:signal_h);

use sigtrap 'handler' => \&signal_term, 'INT';
use sigtrap 'handler' => \&signal_term, 'TERM';
use sigtrap 'handler' => \&signal_term, 'KILL';
use sigtrap 'handler' => \&signal_usr1, 'USR1';
use sigtrap 'handler' => \&signal_usr2, 'USR2';

use lib_stdlog;
lib_stdlog::set_stdlog($0, 5120000);
use prontus_varglb;
prontus_varglb::init();

use FCGI;
use FCGI::ProcManager;

# Modulos usados por las bibliotecas Prontus incluidas abajo.
use DBI;
use Digest::MD5;
use glib_dbi_03;
use glib_fildir_03;
use lib_friendly;
use lib_utils;
use lib_prontus;
use lib_lang;
use lib_bd;
use lib_urlf5;

use Data::Dumper;

my $DEBUG = 0;
my $PROC_MANAGER;
my $SOCKET;
my $REQUEST;
my $SCRIPT_FILE = $0;
my $LOAD_TIME = 0;
my $LASTINIT = 0;
my $DIN_TAX = 0;
my $PRONTUS_MV_ID = '';
my $SIGSET;
my $URL_CACHE_SIZE = 100000;
my $USE_BD = 0;

# estructura del hash
# $URLS{$uri} => {
#    'access' => $epoch,
#    'path' => $path,
#    'res' => $res
#    }
#
# estructura del hash con bd
# $URLS{$uri} => {
#    'access' => 1686290476,
#    'urlf5_file' => '/next/site/tax/port/all/taxport_1___1.html',
#    'urlf5_prontus' => 'next',
#    'urlf5_type' => 'tax',
#    'urlf5_subtypeid' => 1,
#    'urlf5_url' => '/prefijo/tpaises',
#    'urlf5_subtype' => 'seccion',
#    'urlf5_id' => 113
#    }

my %URLS;
my %URL_MULTITAG;
my $MTAG_CACHE_DIR;
my $CACHE_SIZE = 0;

main: {
    my %conf = %{lib_utils::getHash("$prontus_varglb::DIR_SERVER/$prontus_varglb::DIR_CGI_CPAN/fcgi_fpm/fcgi.cfg")};

    my $numprocs = $prontus_varglb::FCGI_CONFIG{'num_procs_friendly'};
    # Asigna valor de configuracion si existe
    if (defined($conf{'num_procs_friendly'})) {
        $numprocs = $conf{'num_procs_friendly'};
    }

    if (defined($conf{'debug'})) {
        $DEBUG = $conf{'debug'};
    }

    if (defined($conf{'use_db'})) {
        $USE_BD = $conf{'use_db'};
    }

    if (defined($conf{'url_cache_size'})) {
        $URL_CACHE_SIZE = $conf{'url_cache_size'};
    }

    if (defined($conf{'prontus_mv_conf'})) {
        $PRONTUS_MV_ID = $conf{'prontus_mv_conf'};
    }

    # Busca prontuses y utliza el primero que encuentra (busca en orden alfabetico).
    my @prontus_dirs = lib_prontus::get_prontus_dirs();

    if ($PRONTUS_MV_ID eq '') {
        foreach my $dir (sort @prontus_dirs) {
            if (-f "$prontus_varglb::DIR_SERVER/$dir/cpan/$dir.cfg") {
                $PRONTUS_MV_ID = $dir;
                last;
            }
        }
    }

    &init_conf($PRONTUS_MV_ID);

    $MTAG_CACHE_DIR = "/$PRONTUS_MV_ID/site$prontus_varglb::DIR_MULTITAG/";

    $LASTINIT = time;

    # Construye el directorio del socket.
    my $sockdir = $prontus_varglb::DIR_SERVER;
    $sockdir =~ s|/web.*|/run|;

    $SOCKET = FCGI::OpenSocket( "$sockdir/route_friendly.socket", 10 );
    $REQUEST = FCGI::Request( \*STDIN, \*STDOUT, \*STDERR, \%ENV, $SOCKET, &FCGI::FAIL_ACCEPT_ON_INTR );

    my $procName = $prontus_varglb::FCGI_CONFIG{'f5_proc_name'};
    my $workerName = $prontus_varglb::FCGI_CONFIG{'f5_worker_name'};
    # Asigna valor de configuracion si existe
    if (defined($conf{'procs_id'})) {
        $procName = $prontus_varglb::FCGI_CONFIG{'f5_proc_name'} . '-' .$conf{'procs_id'};
        $workerName = $prontus_varglb::FCGI_CONFIG{'f5_worker_name'} . '-' .$conf{'procs_id'};
    }

    $prontus_varglb::FCGI_MODE = 1;

    $PROC_MANAGER = FCGI::ProcManager->new({
        n_processes => $numprocs,
        pm_title => $procName
    });

    $PROC_MANAGER->pm_manage();
    $PROC_MANAGER->pm_change_process_name($workerName);

    $SIGSET = POSIX::SigSet->new(SIGUSR1);
    $SIGSET->addset(SIGUSR2);

    &handle_requests();
}

sub handle_requests {
    lib_stdlog::print_err("F5_FPM START\n") if ($DEBUG);
    my ($bd, $msg_err_bd);
    if ($USE_BD) {
        # Carga variables de configuracion.
        my $msg_load = lib_utils::load_config_by_id($PRONTUS_MV_ID);
        if ($msg_load ne '') {
            lib_stdlog::print_err($msg_load);
            exit;
        }
        my @vistas = keys %{$prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'MULTIVISTA'}};
        lib_friendly::init($prontus_varglb::DIR_SERVER,$prontus_varglb::DIR_FRIENDLY5,\@vistas);
        # Conectar a BD
        ($bd, $msg_err_bd) = lib_bd::conectar_prontus_bd();
        if (!ref($bd)) {
            lib_stdlog::print_err($msg_err_bd);
            exit;
        }
    }

    while ($REQUEST->Accept() >= 0) {
        $PROC_MANAGER->pm_pre_dispatch();
        sigprocmask(SIG_BLOCK, $SIGSET);

        lib_stdlog::set_stdlog($SCRIPT_FILE, 5120000);

        my $uri = $ENV{'DOCUMENT_URI'};
        my $query_string = $ENV{'QUERY_STRING'};

        lib_stdlog::print_err("\n\nSTART: route_friendly [$uri][$query_string]\n") if ($DEBUG);

        # url a minusculas
        my $uri_norm = lc($uri);
        # elimina / repetidos
        $uri_norm =~ s|[/\\]+|/|g;
        # quita caracteres no válidos
        $uri_norm =~ s|[^\w\-/\.]+||g;
        # normalizacion urls a minusculas y quita / final
        my $last_char = substr($uri, -1);
        if ($uri_norm ne $uri || $last_char eq '/') {
            if ($last_char eq '/') {
                $uri_norm = substr($uri_norm, 0, -1);
            }
            if ($query_string) {
                $uri_norm .= '?'.$query_string;
            }
            &print_response(301, $uri_norm);
        }
        # procesar multitag
        elsif (index($uri, 'seccion/') > -1 or index($uri, 'tema/') > -1 or index($uri, 'subtema/') > -1) {
            my $md5_url = Digest::MD5::md5_hex($uri);
            my $file = $prontus_varglb::DIR_SERVER . $MTAG_CACHE_DIR . "$md5_url.html";

            # el archivo existe
            if (-f $file) {
                # md5 url en hash, fue vistiado antes, revisamos si no expiro
                if (defined($URL_MULTITAG{$md5_url}) &&
                    (time - $URL_MULTITAG{$md5_url}) > 300) {
                    # archivo existe y no esta expirado
                    &print_response(200, $MTAG_CACHE_DIR . "$md5_url.html");
                }
                # url no ha sido visitada, hay que revisar si esta expirada
                elsif (!defined($URL_MULTITAG{$md5_url})) {
                    my $modtime = (stat $file)[9];
                    # menor a 300 segundos de antiguedad es url valida
                    if ((time - $modtime) < 300) {
                        # guardamos el valor de modificacion para otro ciclo
                        $URL_MULTITAG{$md5_url} = $modtime;
                        &print_response(200, $MTAG_CACHE_DIR . "$md5_url.html");
                    }
                    # en caso contrario el proceso continua para llamar a la cgi y refrescar la página
                }
            }
            # el proceso continua por aca si:
            # la página no ha sido visitada, expiró o el archivo no existe
            # se llamar a la cgi para generar la página

            # por defecto se usa el prontus configurado com principal
            my $query = "?prontus_id=$PRONTUS_MV_ID";
            # revisa si el primer match contiene prontus id
            if ($uri =~ m|^/([^\/]+)|) {
                if ($1 ne 'seccion' && $1 ne 'tema' && $1 ne 'subtema') {
                    $query = "?prontus_id=$1";
                }
            }
            if ($uri =~ m|/seccion/([^\/]+)|) {
                $query .= "&s=$1";
            }
            if ($uri =~ m|/tema/([^\/]+)|) {
                $query .= "&t=$1";
            }
            if ($uri =~ m|/subtema/([^\/]+)|) {
                $query .= "&st=$1";
            }
            if ($uri =~ m|/p/(\d+)|) {
                $query .= "&p=$1";
            }

            $URL_MULTITAG{$md5_url} = time;
            &print_response(200, "/cgi-bin/multitag/prontus_multitag_list.cgi$query");
        }
        # generar taxport dinamicas
        elsif (index($uri, '/site/tax/port') > -1) {
            lib_stdlog::print_err("Taxport url [$uri] DIN_TAX[$DIN_TAX\n") if ($DEBUG);
            # procesar friendly url
            my ($res, $path);
            if ($USE_BD) {
                ($res, $path) = &resolve_url_db($bd, $uri, $query_string);
            } else {
                ($res, $path) = &resolve_url_link($uri, $query_string);
            }

            if ($res != 404) {
                &print_response($res, $path);
            } else {
                # si estan activas se generan en caso contrario no hace nada
                if ($DIN_TAX) {
                    my $cmd = "/usr/bin/perl $prontus_varglb::DIR_SERVER/cgi-bin/prontus_make_taxport.pl $uri";
                    system $cmd;
                    if (-f "$prontus_varglb::DIR_SERVER$uri") {
                        if ($query_string) {
                            $uri .= '?'.$query_string;
                        }
                        &print_response(200, $uri);
                    } else {
                        &print_response(404);
                    }
                }
                # si no hay taxport dinamica se retorna 404, el archivo no fue encontrado por el servidor previamente
                else {
                    &print_response(404);
                }
            }
        }
        # procesar friendly url
        elsif ($USE_BD) {
            my ($res, $path) = &resolve_url_db($bd, $uri, $query_string);
            &print_response($res, $path);
        } else {
            my ($res, $path) = &resolve_url_link($uri, $query_string);
            &print_response($res, $path);
        }

        sigprocmask(SIG_UNBLOCK, $SIGSET);
        $PROC_MANAGER->pm_post_dispatch();
    }

    FCGI::CloseSocket($SOCKET);
}

sub print_response {
    my ($res, $path) = @_;
    if ($res == 200) {
        print "X-Accel-Redirect: $path\n\n";
    } elsif ($res == 301) {
        print "Status: 301 Moved Permanently\nLocation: $path\n\n";
    } elsif ($res == 302) {
        print "Status: 302 Found\nLocation: $path\n\n";
    } elsif ($res == 406) {
        print "Status: 406 Not Acceptable\n\n";
    } else{
        print "Status: 404 Not Found\n\n";
    }
}

sub resolve_url_db {
    my ($bd, $uri, $query_string) = @_;
    my $pag;
    # Detecta si se esta solicitando un friendly con vista, separando
    # la vista del friendly sin vista.
    my ($vista, $friendly) = lib_friendly::getVista($uri);

    # Detecta si se esta solicitando un friendly con pagina, separando
    # la pagina del friendly sin pagina.
    ($pag, $friendly) = lib_friendly::getPag($friendly);

    my $url_data;
    if (defined($URLS{$friendly})) {
        $url_data = $URLS{$friendly};
        $URLS{$friendly}{'access'} = time;
    } else {
        $url_data = lib_urlf5::get_url_data($bd, $friendly);
        if ($url_data) {
            $URLS{$friendly} = $url_data;
            $URLS{$friendly}{'access'} = time;
            $CACHE_SIZE++;
            &garbage_cache() if ($CACHE_SIZE > $URL_CACHE_SIZE );
        }
    }

    my $path = $url_data->{'urlf5_file'};
    if ($query_string) {
        $path .= '?'.$query_string;
    }
    # no hay info de esta uri
    if (!$url_data->{'urlf5_file'}) {
        return (404, '');
    }
    # es un redirect 301
    elsif ($url_data->{'urlf5_subtype'} eq '301') {
        return (301, $path);
    }
    # es un redirect 302
    elsif ($url_data->{'urlf5_subtype'} eq '302') {
        return (302, $path);
    }
    # es un "rewrite" de un archivo fisico, que no es portada, articulo, tax o tag
    elsif ($url_data->{'urlf5_subtype'} eq '200') {
        return (200, $path);
    }
    # el resto de las urls normal
    # si hay taxport dinamicas o no existe se generara otro llamado que si la genera
    # evita chequear si existe adicionalmente
    else {
        my $composedPath = lib_friendly::insertVista($vista, $url_data->{'urlf5_file'});
        if ($pag) {
            ($composedPath) = lib_friendly::insertPag($pag, $composedPath);
        }

        if ($query_string) {
            $composedPath .= '?'.$query_string;
        }

        return (200, $composedPath);
    }
}

sub resolve_url_link {
    my ($uri, $query_string) = @_;
    # Inicializa (de nuevo) si han pasado mas de 60 segundos.
    if (time - $LASTINIT > 60) {
        &init_conf($PRONTUS_MV_ID);
        $LASTINIT = time;
    }
    my ($res, $path, $msg) = ('', '', '');
    if (defined($URLS{$uri})) {
        $res = $URLS{$uri}{'res'};
        $path = $URLS{$uri}{'path'};
        $URLS{$uri}{'access'} = time;
    } else {
        ($res, $path, $msg) = lib_friendly::readFriendly($uri);
        if ($res == 1 || $res == 6 || $res == 7) {
            # lib_stdlog::print_err(Dumper(\%URLS));
            lib_stdlog::print_err("add url to cache 'path' => $path, 'res' => $res\n") if ($DEBUG);
            $URLS{$uri} = { 'access' => time, 'path' => $path, 'res' => $res };
            $CACHE_SIZE++;
            &garbage_cache() if ($CACHE_SIZE > $URL_CACHE_SIZE );
        }
    }

    if ($DEBUG) {
        lib_stdlog::print_err("Result route_friendly [$prontus_varglb::DIR_SERVER][$uri][$prontus_varglb::DIR_FRIENDLY5] [$res, $path, $msg]\n");
    }

    if ($query_string) {
        $path .= '?'.$query_string;
    }

    if ($res == 1) {
        return (200, $path);
    } elsif ($res == 6) {
        return (301, $path);
    } elsif ($res == 7) {
        return (302, $path);
    } else{
        return (404, '');
    }
}

# Inicializacion de la configuración
sub init_conf {
    my $prontus_id = $_[0];
    # Lee las vistas desde la configuracion de Prontus.
    my @vistas = ();
    my $cfgFile = "$prontus_varglb::DIR_SERVER/$prontus_id/cpan/$prontus_id.cfg";
    lib_stdlog::print_err("init cfg: [$cfgFile]\n") if ($DEBUG);

    my $mtime = (stat $cfgFile)[9];
    if ($LOAD_TIME < $mtime) {
        lib_stdlog::print_err("Load conf\n") if ($DEBUG);
        $LOAD_TIME = $mtime;
        my $buffer = glib_fildir_03::read_file($cfgFile);
        while($buffer =~ /MULTIVISTA\s*=\s*["']([^'"]+)["']/g) {
            lib_stdlog::print_err("Vista [$1]\n") if ($DEBUG);
            push @vistas, $1;
        }

        if ($buffer =~ /TAXPORT_DINAMICAS\s*=\s*["']SI["']/) {
            $DIN_TAX = 1;
        }
        lib_stdlog::print_err("DIN_TAX [$DIN_TAX\n") if ($DEBUG);

        lib_friendly::init($prontus_varglb::DIR_SERVER,$prontus_varglb::DIR_FRIENDLY5,\@vistas);
    }
} # init

sub garbage_cache {
    # lib_stdlog::print_err("garbage_cache\n");
    # lib_stdlog::print_err(Dumper(\%URLS));

    my $counter = 0;
    foreach my $uri (sort { $URLS{$b}{'access'} <=> $URLS{$a}{'access'} } keys %URLS) {
        # lib_stdlog::print_err("$uri  -". $URLS{$uri}{'access'}."\n");
        $counter++;
        if ($counter > $URL_CACHE_SIZE) {
            delete $URLS{$uri};
        }
    }
}

sub signal_usr1 {
    lib_stdlog::print_err("[$$] Reload config\n");
    if (!$USE_BD) {
        # recarga configuracion de multivistas para friendly
        $LOAD_TIME = 0;
        &init_conf($PRONTUS_MV_ID);
    } else {
        # limpia variables de configuracion para que sean recargadas
        lib_prontus::reset_prontus_config();
    }
    # desbloquea señal para recibirla nuevamente
    sigprocmask(SIG_UNBLOCK, $SIGSET);
    # reinicia proceso
    &handle_requests();
}

sub signal_usr2 {
    lib_stdlog::print_err("[$$] Reset friendly url cache\n");
    # borra cache de friendlys
    %URLS = ();
    # desbloquea señal para recibirla nuevamente
    sigprocmask(SIG_UNBLOCK, $SIGSET);
    # reinicia proceso
    &handle_requests();
}

# funcion para capturar las señales INT y TERM y logearlas
sub signal_term {
    lib_stdlog::print_err("STDERR: [$$] Ended by signal", @_ , "\n");
    $PROC_MANAGER->pm_exit();
    exit(0);
}
