#!/usr/bin/perl

# --------------------------------------------------------------
# Prontus CMS
# http://www.prontus.cl
# by Altavoz.net
#
# licensed under LGPL license.
# http://www.prontus.cl/license.html
# --------------------------------------------------------------
#
#-------------------------------COMENTARIO GLOBAL---------------
#---------------------------------------------------------------
# PROPOSITO.
#-----------
# Subrutinas para localización e internacionalización de Prontus
#---------------------------------------------------------------

#---------------------------------------------------------------
# DECLARACIONES GLOBALES.
#------------------------

package lib_lang;
use strict;
use utf8;

our %TRANSLATION_TABLE = (); # Contendrá la lista de strings disponibles.
our $LANGUAGE = '';
 # Idiomas disponible
our %AVAIL_LANGUAGE = ();

# devuelve un array de hashes con el listado de idiomas disponibles para localización de Prontus.
sub get_available_languages {
    if (scalar keys %AVAIL_LANGUAGE > 0) {
        return \%AVAIL_LANGUAGE;
    }

    my $basedir = "$prontus_varglb::DIR_SERVER/prontus/cpan/lang";
    my $filename = "$basedir/languages.json";

    # archivo no legible
    if (! -r $filename) {
        return \%AVAIL_LANGUAGE;
    }

    my $strjson = glib_fildir_03::read_file_bin($filename);
    my $languages = eval { JSON::decode_json($strjson) };
    if ($@) {
        lib_stdlog::print_err(lib_lang::get_interpolated_message('global_json_decode_error', $@, $strjson)."\n");
        return \%AVAIL_LANGUAGE;
    }

    # procesamos la lista de idiomas
    foreach my $code (keys %{$languages}) {
        $AVAIL_LANGUAGE{$code} = $languages->{$code};
    }
    return \%AVAIL_LANGUAGE;
}

# devuelve el lenguaje seleccionado, si es válido (existe un archivo de lenguaje), o el lenguaje por defecto.
sub get_language {
    my $lang = $_[0];
    &get_available_languages();

    if (exists($AVAIL_LANGUAGE{$lang})) {
        return $lang;
    } else {
        # si todo lo demás falla, devolvemos español.
        return $prontus_varglb::CONFIG{$prontus_varglb::PRONTUS_ID}{'DEFAULT_LANGUAGE'} || $prontus_varglb::DEFAULT_LANGUAGE;
    }
}

# devuelve el lenguaje seleccionado configurado para el usuario, si es válido (existe un archivo de lenguaje), o el lenguaje por defecto.
sub get_user_language {
    $prontus_auth::USERS_LANG = &get_language($prontus_auth::USERS_LANG);
    return $prontus_auth::USERS_LANG;
}

# lee la tabla de traducciones de Prontus
# genera un hash de hashes:
# TRANSLATION_TABLE = (
#   es => {
#       key => mensaje,
#       key2 => mensaje2,
#       ...
#   },
#   en => {
#         key => message,
#         key2 => message2,
#         ...
#   },
# );
# @return un hash con la tabla de traducción para un idioma, o vacío en caso de error.
sub load_translation_table {
    my $lang = $_[0];
    $LANGUAGE = $lang;
    #Si ya hemos cargado la tabla de traducción de este idioma, retornamos.
    if ($TRANSLATION_TABLE{$lang}) {
        return;
    }

    # de otro modo, la cargamos y retornamos.
    my $lang_file = $prontus_varglb::DIR_SERVER. "/prontus/cpan/lang/$lang/l10n.lang";
    if (! -s $lang_file) {
        lib_stdlog::print_err("Translation file [$lang_file] not found.\n");
        return;
    }
    $TRANSLATION_TABLE{$lang} = ();

    my $buffer = glib_fildir_03::read_file($lang_file);
    $buffer =~ s/^\s+|\s+$//g;
    return if $buffer eq '';

    my %tmp_hash = &process_buffer($buffer);
    for my $tmp_key (keys %tmp_hash) {
        $TRANSLATION_TABLE{$lang}{$tmp_key} = $tmp_hash{$tmp_key};
    }
}

# recibe el texto de un archivo de traducción, devuelve un hash con la tabla de traducción.
# @param string $buffer
# @return hash
sub process_buffer {
    my $buffer = $_[0];
    $buffer =~ s/\r/\n/g;
    $buffer =~ s/\n{2,}/\n/g;
    my @lines = split(/\n/, $buffer);
    my %table;
    for my $line (@lines) {
        $line =~ s/^\s+|\s+$//g;
        next if $line eq ''; # línea vacía
        next if index($line, ';') == 0; # comentario en formato de .ini
        next if index($line, '[') == 0; # section
        next if $line !~ /^\w+=.*/; # el formato de línea es [id alfanumérico con _]=[cualquier cosa]
        # https://perldoc.perl.org/functions/split
        # If LIMIT is specified and positive, it represents the maximum number of fields
        # into which the EXPR may be split; in other words, LIMIT is one greater than the
        # maximum number of times EXPR may be split.
        #Thus, the LIMIT value 1 means that EXPR may be split a maximum of zero times
        my ($key, $val) = split(/=/, $line, 2); #formato identificador=valor
        $val =~ s/\\n/\n/g;
        $val =~ s/\\t/\t/g;
        $table{$key} = $val;
    }
    return %table;
}

# devuelve la traducción de una string desde la translation table, o la key si no se encuentra traducción.
sub get_message {
    my $key = $_[0];
    return $TRANSLATION_TABLE{$LANGUAGE}{$key} || $key;
}

# devuelve la traducción de una string desde la translation table, o la key si no se encuentra traducción.
# interpola eventuales argumentos sucesivos en la string.
sub get_interpolated_message {
    my ($key, @args) = @_;
    return $key if !$TRANSLATION_TABLE{$LANGUAGE}{$key};
    return sprintf($TRANSLATION_TABLE{$LANGUAGE}{$key}, @args);
}

1;
